from menpo.io.landmark import PTSImporter
import numpy as np


class MeshPTSImporter(PTSImporter):
    r"""
    Implements the :meth:`_build_points` method for meshes. Here, ``x`` is the
    first axis.

    Parameters
    ----------
    filepath : string
        Absolute filepath of the landmarks
    """

    def __init__(self, filepath):
        super(MeshPTSImporter, self).__init__(filepath)

    def _build_points(self, xs, ys):
        """
        For meshes, ``axis 0 = xs`` and ``axis 1 = ys``. Therefore, return the
        appropriate points array ordering.

        Parameters
        ----------
        xs : (N,) ndarray
            Row vector of ``x`` coordinates
        ys : (N,) ndarray
            Row vector of ``y`` coordinates

        Returns
        -------
        points : (N, 2) ndarray
            Array with ``xs`` as the first axis: ``[xs; ys]``
        """
        return np.hstack([xs, ys])



import abc


class Invertible(object):
    r"""
    Transform Mixin for invertible transforms. Provides an interface for
    taking the psuedo or true inverse of a transform.
    """

    @abc.abstractproperty
    def has_true_inverse(self):
        r"""
        True if the pseudoinverse is an exact inverse.

        :type: Boolean
        """
        pass

    @property
    def pseudoinverse(self):
        r"""
        The pseudoinverse of the transform - that is, the transform that
        results from swapping source and target, or more formally, negating
        the transforms parameters. If the transform has a true inverse this
        is returned instead.

        :type: :class:`Transform`
        """
        return self._build_pseudoinverse()

    @abc.abstractmethod
    def _build_pseudoinverse(self):
        r"""
        Returns this transform's inverse if it has one. if not,
        the pseduoinverse is given.

        This method is called by the pseudoinverse property and must be
        overridden.


        Returns
        -------
        pseudoinverse: type(self)
        """
        pass


class VInvertible(Invertible):
    r"""
    Transform Mixin for Vectorizable invertible transforms.

    Prefer this Mixin over Invertible if the Transform in question is
    Vectorizable as this adds from_vector variants to the Invertible
    interface. These can be tuned for performance, and are for instance
    needed by some of the machinery of AAMs.
    """

    def pseudoinverse_vector(self, vector):
        r"""
        The vectorized pseudoinverse of a provided vector instance.

        Syntactic sugar for

        self.from_vector(vector).pseudoinverse.as_vector()

        Can be much faster than the explict call as object creation can be
        entirely avoided in some cases.

        Parameters
        ----------
        vector :  (P,) ndarray
            A vectorized version of self

        Returns
        -------
        pseudoinverse_vector : (N,) ndarray
            The pseudoinverse of the vector provided
        """
        return self.from_vector(vector).pseudoinverse.as_vector()



from copy import deepcopy
import numpy as np
from menpo.image.base import Image
from skimage.transform import pyramid_gaussian


class BooleanImage(Image):
    r"""
    A mask image made from binary pixels. The region of the image that is
    left exposed by the mask is referred to as the 'masked region'. The
    set of 'masked' pixels is those pixels corresponding to a True value in
    the mask.

    Parameters
    -----------
    mask_data : (M, N, ..., L) ndarray
        The binary mask data. Note that there is no channel axis - a 2D Mask
         Image is built from just a 2D numpy array of mask_data.
        Automatically coerced in to boolean values.
    """

    def __init__(self, mask_data):
        # Enforce boolean pixels, and add a channel dim
        mask_data = np.asarray(mask_data[..., None], dtype=np.bool)
        super(BooleanImage, self).__init__(mask_data)

    @classmethod
    def _init_with_channel(cls, image_data_with_channel):
        r"""
        Constructor that always requires the image has a
        channel on the last axis. Only used by from_vector. By default,
        just calls the constructor. Subclasses with constructors that don't
        require channel axes need to overwrite this.
        """
        return cls(image_data_with_channel[..., 0])

    @classmethod
    def blank(cls, shape, fill=True, round='ceil', **kwargs):
        r"""
        Returns a blank :class:`BooleanImage` of the requested shape

        Parameters
        ----------
        shape : tuple or list
            The shape of the image. Any floating point values are rounded
            according to the ``round`` kwarg.

        fill : True or False, optional
            The mask value to be set everywhere

            Default: True (masked region is the whole image - meaning the whole
                 image is exposed)
        round: {'ceil', 'floor', 'round'}
            Rounding function to be applied to floating point shapes.

            Default: 'ceil'

        Returns
        -------
        blank_image : :class:`BooleanImage`
            A blank mask of the requested size
        """
        if round not in ['ceil', 'round', 'floor']:
            raise ValueError('round must be either ceil, round or floor')
            # Ensure that the '+' operator means concatenate tuples
        shape = tuple(getattr(np, round)(shape))
        if fill:
            mask = np.ones(shape, dtype=np.bool)
        else:
            mask = np.zeros(shape, dtype=np.bool)
        return cls(mask)

    @property
    def mask(self):
        r"""
        Returns the pixels of the mask with no channel axis. This is what
        should be used to mask any k-dimensional image.

        :type: (M, N, ..., L), np.bool ndarray
        """
        return self.pixels[..., 0]

    @property
    def n_true(self):
        r"""
        The number of ``True`` values in the mask

        :type: int
        """
        return np.sum(self.pixels)

    @property
    def n_false(self):
        r"""
        The number of ``False`` values in the mask

        :type: int
        """
        return self.n_pixels - self.n_true

    @property
    def proportion_true(self):
        r"""
        The proportion of the mask which is ``True``

        :type: double
        """
        return (self.n_true * 1.0) / self.n_pixels

    @property
    def proportion_false(self):
        r"""
        The proportion of the mask which is ``False``

        :type: double
        """
        return (self.n_false * 1.0) / self.n_pixels

    @property
    def true_indices(self):
        r"""
        The indices of pixels that are true.

        :type: (``n_dims``, ``n_true``) ndarray
        """
        # Ignore the channel axis
        return np.vstack(np.nonzero(self.pixels[..., 0])).T

    @property
    def false_indices(self):
        r"""
        The indices of pixels that are false.

        :type: (``n_dims``, ``n_false``) ndarray
        """
        # Ignore the channel axis
        return np.vstack(np.nonzero(~self.pixels[..., 0])).T

    @property
    def all_indices(self):
        r"""
        Indices into all pixels of the mask, as consistent with
        true_indices and false_indices

        :type: (``n_dims``, ``n_pixels``) ndarray
        """
        return np.indices(self.shape).reshape([self.n_dims, -1]).T

    def __str__(self):
        return ('{} {}D mask, {:.1%} '
                'of which is True '.format(self._str_shape, self.n_dims,
                                           self.proportion_true))

    def from_vector(self, flattened):
        r"""
        Takes a flattened vector and returns a new
        :class:`BooleanImage` formed by
        reshaping the vector to the correct dimensions. Note that this is
        rebuilding a boolean image **itself** from boolean values. The mask
        is in no way interpreted in performing the operation, in contrast to
        MaskedImage, where only the masked region is used in from_vector()
        and as_vector(). Any image landmarks are transferred in the process.

        Parameters
        ----------
        flattened : (``n_pixels``,) np.bool ndarray
            A flattened vector of all the pixels of a BooleanImage.

        Returns
        -------
        image : :class:`BooleanImage`
            New BooleanImage of same shape as this image
        """
        mask = BooleanImage(flattened.reshape(self.shape))
        mask.landmarks = self.landmarks
        return mask

    def invert(self):
        r"""
        Inverts the current mask in place, setting all True values to False,
        and all False values to True.
        """
        self.pixels = ~self.pixels

    def inverted_copy(self):
        r"""
        Returns a copy of this Boolean image, which is inverted.

        Returns
        -------
        inverted_image: :class:`BooleanNSImage`
            An inverted copy of this boolean image.
        """
        inverse = deepcopy(self)
        inverse.invert()
        return inverse

    def bounds_true(self, boundary=0, constrain_to_bounds=True):
        r"""
        Returns the minimum to maximum indices along all dimensions that the
        mask includes which fully surround the True mask values. In the case
        of a 2D Image for instance, the min and max define two corners of a
        rectangle bounding the True pixel values.

        Parameters
        ----------
        boundary : int, optional
            A number of pixels that should be added to the extent. A
            negative value can be used to shrink the bounds in.

            Default: 0

        constrain_to_bounds: bool, optional
            If True, the bounding extent is snapped to not go beyond
            the edge of the image. If False, the bounds are left unchanged.

            Default: True

        Returns
        --------
        min_b : (D,) ndarray
            The minimum extent of the True mask region with the boundary
            along each dimension. If constrain_to_bounds was True,
            is clipped to legal image bounds.

        max_b : (D,) ndarray
            The maximum extent of the True mask region with the boundary
            along each dimension. If constrain_to_bounds was True,
            is clipped to legal image bounds.
        """
        mpi = self.true_indices
        maxes = np.max(mpi, axis=0) + boundary
        mins = np.min(mpi, axis=0) - boundary
        if constrain_to_bounds:
            maxes = self.constrain_points_to_bounds(maxes)
            mins = self.constrain_points_to_bounds(mins)
        return mins, maxes

    def bounds_false(self, boundary=0, constrain_to_bounds=True):
        r"""
        Returns the minimum to maximum indices along all dimensions that the
        mask includes which fully surround the False mask values. In the case
        of a 2D Image for instance, the min and max define two corners of a
        rectangle bounding the False pixel values.

        Parameters
        ----------
        boundary : int >= 0, optional
            A number of pixels that should be added to the extent. A
            negative value can be used to shrink the bounds in.

            Default: 0

        constrain_to_bounds: bool, optional
            If True, the bounding extent is snapped to not go beyond
            the edge of the image. If False, the bounds are left unchanged.

            Default: True

        Returns
        --------
        min_b : (D,) ndarray
            The minimum extent of the False mask region with the boundary
            along each dimension. If constrain_to_bounds was True,
            is clipped to legal image bounds.

        max_b : (D,) ndarray
            The maximum extent of the False mask region with the boundary
            along each dimension. If constrain_to_bounds was True,
            is clipped to legal image bounds.
        """
        return self.inverted_copy().bounds_true(
            boundary=boundary, constrain_to_bounds=constrain_to_bounds)

    def warp_to(self, template_mask, transform, warp_landmarks=False,
                interpolator='scipy', **kwargs):
        r"""
        Warps this BooleanImage into a different reference space.

        Parameters
        ----------
        template_mask : :class:`menpo.image.boolean.BooleanImage`
            Defines the shape of the result, and what pixels should be
            sampled.
        transform : :class:`menpo.transform.base.Transform`
            Transform **from the template space back to this image**.
            Defines, for each True pixel location on the template, which pixel
            location should be sampled from on this image.
        warp_landmarks : bool, optional
            If ``True``, warped_image will have the same landmark dictionary
            as self, but with each landmark updated to the warped position.

            Default: ``False``
        interpolator : 'scipy' or 'c', optional
            The interpolator that should be used to perform the warp.

            Default: 'scipy'
        kwargs : dict
            Passed through to the interpolator. See `menpo.interpolation`
            for details.

        Returns
        -------
        warped_image : type(self)
            A copy of this image, warped.
        """
        # enforce the order as 0, for this boolean data, then call super
        manually_set_order = kwargs.get('order', 0)
        if manually_set_order != 0:
            raise ValueError(
                "The order of the interpolation on a boolean image has to be "
                "0 (attempted to set {})".format(manually_set_order))
        kwargs['order'] = 0
        return Image.warp_to(self, template_mask, transform,
                             warp_landmarks=warp_landmarks,
                             interpolator=interpolator, **kwargs)

    def _build_warped_image(self, template_mask, sampled_pixel_values,
                            **kwargs):
        r"""
        Builds the warped image from the template mask and
        sampled pixel values. Overridden for BooleanImage as we can't use
        the usual from_vector_inplace method.
        """
        warped_image = BooleanImage.blank(template_mask.shape)
        # As we are a mask image, we have to implement the update a little
        # more manually than other image classes.
        warped_image.pixels[warped_image.mask] = sampled_pixel_values
        return warped_image



import numpy as np
from scipy.linalg.blas import dgemm
from menpo.decomposition import principal_component_decomposition
from menpo.model.base import MeanInstanceLinearModel


class PCAModel(MeanInstanceLinearModel):
    """
    A :class:`MeanLinearInstanceModel` where the components are Principal
    Components.

    Principal Component Analysis (PCA) by Eigenvalue Decomposition of the
    data's scatter matrix.

    For details of the implementation of PCA, see :func:`menpo.decomposition
    .principal_component_decomposition`.

    Parameters
    ----------
    samples: list of :class:`menpo.base.Vectorizable`
        List of samples to build the model from.

    center : bool, optional
        When True (True by default) PCA is performed after mean centering the
        data. If False the data is assumed to be centred, and the mean will
        be 0.

    bias: bool, optional
        When True (False by default) a biased estimator of the covariance
        matrix is used, i.e.:

            \frac{1}{N} \sum_i^N \mathbf{x}_i \mathbf{x}_i^T

        instead of default:

            \frac{1}{N-1} \sum_i^N \mathbf{x}_i \mathbf{x}_i^T
    """
    def __init__(self, samples, center=True, bias=False):
        self.samples = samples
        self.center = center
        self.bias = bias
        # build data matrix
        n_samples = len(samples)
        n_features = samples[0].n_parameters
        data = np.zeros((n_samples, n_features))
        for i, sample in enumerate(samples):
            data[i] = sample.as_vector()

        eigenvectors, eigenvalues, mean_vector = \
            principal_component_decomposition(data, whiten=False,
                                              center=center, bias=bias)
        self._eigenvalues = eigenvalues

        super(PCAModel, self).__init__(eigenvectors, mean_vector, samples[0])
        self._n_components = self.n_components
        self._trimmed_variance = 0

    @property
    def n_active_components(self):
        r"""
        The number of components currently in use on this model.
        """
        return self._n_components

    @n_active_components.setter
    def n_active_components(self, value):
        value = round(value)
        if 0 < value <= self.n_components:
            self._n_components = value
        else:
            raise ValueError(
                "Tried setting n_components as {} - has to be an int and "
                "0 < n_components <= n_available_components "
                "(which is {}) ".format(value, self.n_components))

    @MeanInstanceLinearModel.components.getter
    def components(self):
        r"""
        The matrix containing the active components on this model.

        type: (n_active_components, n_features) ndarray
        """
        return self._components[:self.n_active_components, :]

    @property
    def whitened_components(self):
        return self.components / (
            np.sqrt(self.eigenvalues + self.noise_variance)[:, None])

    @property
    def n_samples(self):
        return len(self.samples)

    @property
    def eigenvalues(self):
        return self._eigenvalues[:self.n_active_components]

    @property
    def eigenvalues_ratio(self):
        return self.eigenvalues / self.eigenvalues.sum()

    @property
    def noise_variance(self):
        if self.n_active_components == self.n_components:
            return self._trimmed_variance
        else:
            return (self._eigenvalues[self.n_active_components:].mean() +
                    self._trimmed_variance)

    @property
    def inverse_noise_variance(self):
        noise_variance = self.noise_variance
        if noise_variance == 0:
            raise ValueError("noise variance is nil - cannot take the "
                             "inverse")
        else:
            return 1.0 / noise_variance

    @property
    def jacobian(self):
        """
        Returns the Jacobian of the PCA model. In this case, simply the
        components of the model reshaped to have the standard Jacobian shape:

            n_points    x  n_params      x  n_dims
            n_features  x  n_components  x  n_dims

        Returns
        -------
        jacobian : (n_features, n_components, n_dims) ndarray
            The Jacobian of the model in the standard Jacobian shape.
        """
        jacobian = self.components.reshape(self.n_active_components, -1,
                                           self.template_instance.n_dims)
        return jacobian.swapaxes(0, 1)

    def component_vector(self, index, with_mean=True, scale=1.0):
        r"""
        A particular component of the model, in vectorized form.

        Parameters
        ----------
        index : int
            The component that is to be returned

        with_mean: boolean (optional)
            If True, the component will be blended with the mean vector
            before being returned. If not, the component is returned on it's
            own.

            Default: True
        scale : float
            A scale factor that should be applied to the component. Only
            valid in the case where with_mean is True. The scale is applied
            in units of standard deviations (so a scale of 1.0
            with_mean visualizes the mean plus 1 std. dev of the component
            in question).

        :type: (n_features,) ndarray
        """
        if with_mean:
            # on PCA, scale is in units of std. deviations...
            scaled_eigval = scale * np.sqrt(self.eigenvalues[index])
            return (scaled_eigval * self.components[index]) + self.mean_vector
        else:
            return self.components[index]

    def instance_vectors(self, weights):
        """
        Creates new vectorized instances of the model using the first
        components in a particular weighting.

        Parameters
        ----------
        weights : (n_vectors, n_weights) ndarray or list of lists
            The weightings for the first n_weights components that
            should be used per instance that is to be produced

            ``weights[i, j]`` is the linear contribution of the j'th
            principal component to the i'th instance vector produced. Note
            that if n_weights < n_components, only the first n_weight
            components are used in the reconstruction (i.e. unspecified
            weights are implicitly 0)

        Raises
        ------
        ValueError: If n_weights > n_components

        Returns
        -------
        vectors : (n_vectors, n_features) ndarray
            The instance vectors for the weighting provided.
        """
        weights = np.asarray(weights)  # if eg a list is provided
        n_instances, n_weights = weights.shape
        if n_weights > self.n_active_components:
            raise ValueError(
                "Number of weightings cannot be greater than {}".format(
                    self.n_active_components))
        else:
            full_weights = np.zeros((n_instances, self.n_active_components))
            full_weights[..., :n_weights] = weights
            weights = full_weights
        return self._instance_vectors_for_full_weights(weights)

    def trim_components(self, n_components=None):
        r"""
        Permanently trims the components down to a certain amount.

        This will reduce `n_available_components` down to `n_components`
        (either provided or as currently set), freeing up memory in the
        process.

        Once the model is trimmed, the trimmed components cannot be recovered.

        Parameters
        ----------

        n_components: int, optional
            The number of components that are kept. If None,
            self.n_components is used.
        """
        if n_components is None:
            # by default trim using self.n_components
            n_components = self.n_active_components

        # trim components
        if not n_components < self.n_components:
            raise ValueError(
                "n_components ({}) needs to be less than "
                "n_available_components ({})".format(
                n_components, self.n_components))
        else:
            self._components = self._components[:n_components]
        if self.n_active_components > self.n_components:
            # set n_components if necessary
            self.n_active_components = self.n_components
        # store the amount of variance captured by the discarded components
        self._trimmed_variance = \
            self._eigenvalues[self.n_active_components:].mean()
        # make sure that the eigenvalues are trimmed too
        self._eigenvalues = self._eigenvalues[:self.n_components]

    def distance_to_subspace(self, instance):
        """
        Returns a version of ``instance`` where all the basis of the model
        have been projected out and which has been scaled by the inverse of
        the ``noise_variance``

        Parameters
        ----------
        instance : :class:`menpo.base.Vectorizable`
            A novel instance.

        Returns
        -------
        scaled_projected_out : ``self.instance_class``
            A copy of ``instance``, with all basis of the model projected out
            and scaled by the inverse of the ``noise_variance``.
        """
        vec_instance = self.distance_to_subspace_vector(instance.as_vector())
        return instance.from_vector(vec_instance)

    def distance_to_subspace_vector(self, vector_instance):
        """
        Returns a version of ``instance`` where all the basis of the model
        have been projected out and which has been scaled by the inverse of
        the ``noise_variance``.

        Parameters
        ----------
        vector_instance : (n_features,) ndarray
            A novel vector.

        Returns
        -------
        scaled_projected_out: (n_features,) ndarray
            A copy of ``vector_instance`` with all basis of the model projected
            out and scaled by the inverse of the ``noise_variance``.
        """
        return (self.inverse_noise_variance *
                self.project_out_vectors(vector_instance))

    def project_whitened(self, instance):
        """
        Returns a sheared (non-orthogonal) reconstruction of ``instance``.

        Parameters
        ----------
        instance : :class:`menpo.base.Vectorizable`
            A novel instance.

        Returns
        -------
        sheared_reconstruction : ``self.instance_class``
            A sheared (non-orthogonal) reconstruction of ``instance``.
        """
        vector_instance = self.project_whitened_vector(instance.as_vector())
        return instance.from_vector(vector_instance)

    def project_whitened_vector(self, vector_instance):
        """
        Returns a sheared (non-orthogonal) reconstruction of
        ``vector_instance``.

        Parameters
        ----------
        vector_instance : (n_features,) ndarray
            A novel vector.

        Returns
        -------
        sheared_reconstruction : (n_features,) ndarray
            A sheared (non-orthogonal) reconstruction of ``vector_instance``
        """
        whitened_components = self.whitened_components
        weights = dgemm(alpha=1.0, a=vector_instance.T,
                        b=whitened_components.T, trans_a=True)
        return dgemm(alpha=1.0, a=weights.T, b=whitened_components.T,
                     trans_a=True, trans_b=True)

    def orthonormalize_against_inplace(self, linear_model):
        r"""
        Enforces that the union of this model's components and another are
        both mutually orthonormal.

        Note that the model passed in is guaranteed to not have it's number
        of available components changed. This model, however, may loose some
        dimensionality due to reaching a degenerate state.

        The removed components will always be trimmed from the end of
        components (i.e. the components which capture the least variance).
        If trimming is performed, `n_components` and
        `n_available_components` would be altered - see
        :meth:`trim_components` for details.

        Parameters
        -----------
        linear_model : :class:`LinearModel`
            A second linear model to orthonormalize this against.
        """
        # take the QR decomposition of the model components
        Q = (np.linalg.qr(np.hstack((linear_model._components.T,
                                     self._components.T)))[0]).T
        # the model passed to us went first, so all it's components will
        # survive. Pull them off, and update the other model.
        linear_model.components = Q[:linear_model.n_components, :]
        # it's possible that all of our components didn't survive due to
        # degeneracy. We need to trim our components down before replacing
        # them to ensure the number of components is consistent (otherwise
        # the components setter will complain at us)
        n_available_components = Q.shape[0] - linear_model.n_components
        if n_available_components < self.n_components:
            # oh dear, we've lost some components from the end of our model.
            # call trim_components to update our state.
            self.trim_components(n_components=n_available_components)
        # now we can set our own components with the updated orthogonal ones
        self.components = Q[linear_model.n_components:, :]



import numpy as np

from menpo.base import Targetable, Vectorizable
from menpo.model.pdm import PDM, GlobalPDM, OrthoPDM

from .base import Transform, VComposable, VInvertible


class ModelDrivenTransform(Transform, Targetable, Vectorizable,
                           VComposable, VInvertible):
    r"""
    A transform that couples a traditional landmark-based transform to a
    statistical model such that source points of the alignment transform
    are the points of the model. The weights of the transform are just
    the weights of statistical model.

    If no source is provided, the mean of the model is defined as the
    source landmarks of the transform.

    Parameters
    ----------
    model : :class:`menpo.model.base.StatisticalModel`
        A linear statistical shape model.
    transform_cls : :class:`menpo.transform.AlignableTransform`
        A class of :class:`menpo.transform.base.AlignableTransform`
        The align constructor will be called on this with the source
        and target landmarks. The target is
        set to the points generated from the model using the
        provide weights - the source is either given or set to the
        model's mean.
    source : :class:`menpo.shape.base.PointCloud`
        The source landmarks of the transform. If None, the mean of the model
         is used.

        Default: None

    """
    def __init__(self, model, transform_cls, source=None):
        self.pdm = PDM(model)
        self._cached_points = None
        self.transform = transform_cls(source, self.target)

    @property
    def n_dims(self):
        r"""
        The number of dimensions that the transform supports.

        :type: int
        """
        return self.pdm.n_dims

    def _apply(self, x, **kwargs):
        r"""
        Apply this transform to the given object. Uses the internal transform.

        Parameters
        ----------
        x : (N, D) ndarray or a transformable object
            The object to be transformed.
        kwargs : dict
            Passed through to transforms ``apply_inplace`` method.

        Returns
        --------
        transformed : (N, D) ndarray or object
            The transformed object
        """
        return self.transform._apply(x, **kwargs)

    @property
    def target(self):
        return self.pdm.target

    def _target_setter(self, new_target):
        r"""
        On a new target being set, we need to:

        Parameters
        ----------

        new_target: :class:`PointCloud`
            The new_target that we want to set.
        """
        self.pdm.set_target(new_target)

    def _new_target_from_state(self):
        # We delegate to PDM to handle all our Targetable duties. As a
        # result, *we* never need to call _sync_target_for_state, so we have
        # no need for an implementation of this method. Of course the
        # interface demands it, so the stub is here. Contrast with
        # _target_setter, which is required, because we will have to handle
        # external calls to set_target().
        pass

    def _sync_state_from_target(self):
        # Let the pdm update its state
        self.pdm._sync_state_from_target()
        # and update our transform to the new state
        self.transform.set_target(self.target)

    @property
    def n_parameters(self):
        r"""
        The total number of parameters.

        Simply ``n_weights``.

        :type: int
        """
        return self.pdm.n_parameters

    def as_vector(self):
        r"""
        Return the current weights of this transform - this is the
        just the linear model's weights

        Returns
        -------
        params : (``n_parameters``,) ndarray
            The vector of weights
        """
        return self.pdm.as_vector()

    def from_vector_inplace(self, vector):
        r"""
        Updates the ModelDrivenTransform's state from it's
        vectorized form.
        """
        self.pdm.from_vector_inplace(vector)
        # By here the pdm has updated our target state, we just need to
        # update the transform
        self.transform.set_target(self.target)

    def compose_after_from_vector_inplace(self, delta):
        r"""
        Composes two ModelDrivenTransforms together based on the
        first order approximation proposed by Papandreou and Maragos in [1].

        Parameters
        ----------
        delta : (N,) ndarray
            Vectorized :class:`ModelDrivenTransform` to be applied **before**
            self

        Returns
        --------
        transform : self
            self, updated to the result of the composition


        References
        ----------

        .. [1] G. Papandreou and P. Maragos, "Adaptive and Constrained
               Algorithms for Inverse Compositional Active Appearance Model
               Fitting", CVPR08
        """
        model_jacobian = self.pdm.model.jacobian
        points = self.pdm.model.mean.points
        n_points = self.pdm.model.mean.n_points

        # compute:
        # -> dW/dp when p=0
        # -> dW/dp when p!=0
        # -> dW/dx when p!=0 evaluated at the source landmarks

        # dW/dp when p=0 and when p!=0 are the same and simply given by
        # the Jacobian of the model
        dW_dp_0 = model_jacobian
        dW_dp = dW_dp_0
        # dW_dp_0:  n_points  x     n_params     x  n_dims
        # dW_dp:    n_points  x     n_params     x  n_dims

        dW_dx = self.transform.jacobian_points(points)
        # dW_dx:  n_points  x  n_dims  x  n_dims

        #TODO: Can we do this without splitting across the two dimensions?
        dW_dx_x = dW_dx[:, 0, :].flatten()[..., None]
        dW_dx_y = dW_dx[:, 1, :].flatten()[..., None]
        dW_dp_0_mat = np.reshape(dW_dp_0, (n_points * self.n_dims,
                                           self.n_parameters))
        dW_dx_dW_dp_0 = dW_dp_0_mat * dW_dx_x + dW_dp_0_mat * dW_dx_y
        dW_dx_dW_dp_0 = np.reshape(dW_dx_dW_dp_0,
                                   (n_points, self.n_parameters, self.n_dims))
        # dW_dx:          n_points  x  n_dims    x  n_dims
        # dW_dp_0:        n_points  x  n_params  x  n_dims
        # dW_dx_dW_dp_0:  n_points  x  n_params  x  n_dims

        J = np.einsum('ijk, ilk -> jl', dW_dp, dW_dx_dW_dp_0)
        H = np.einsum('ijk, ilk -> jl', dW_dp, dW_dp)

        Jp = np.linalg.solve(H, J)
        # Jp:  n_params  x  n_params

        self.from_vector_inplace(self.as_vector() + np.dot(Jp, delta))
        return self

    @property
    def has_true_inverse(self):
        return False

    def _build_pseudoinverse(self):
        return self.from_vector(-self.as_vector())

    def pseudoinverse_vector(self, vector):
        r"""
        The vectorized pseudoinverse of a provided vector instance.

        Syntactic sugar for

        self.from_vector(vector).pseudoinverse.as_vector()

        On ModelDrivenTransform this is especially fast - we just negate the
        vector provided.

        Parameters
        ----------
        vector :  (P,) ndarray
            A vectorized version of self

        Returns
        -------
        pseudoinverse_vector : (N,) ndarray
            The pseudoinverse of the vector provided
        """
        return -vector

    def jacobian(self, points):
        """
        Calculates the Jacobian of the ModelDrivenTransform wrt to
        its weights (the weights). This is done by chaining the relative
        weight of each point wrt the source landmarks, i.e. the Jacobian of
        the warp wrt the source landmarks when the target is assumed to be
        equal to the source (dW/dx), together with the Jacobian of the
        linear model  wrt its weights (dX/dp).

        Parameters
        -----------
        points: (N, D) ndarray
            The points at which the Jacobian will be evaluated.

        Returns
        -------
        dW/dp : (N, P, D) ndarray
            The Jacobian of the ModelDrivenTransform evaluated at the
            previous points.
        """
        # check if re-computation of dW/dx can be avoided
        if not np.array_equal(self._cached_points, points):
            # recompute dW/dx, i.e. the relative weight of each point wrt
            # the source landmarks
            self.dW_dX = self.transform.weight_points(points)
            # cache points
            self._cached_points = points

        # dX/dp is simply the Jacobian of the model
        dX_dp = self.pdm.model.jacobian

        # dW_dX:    n_points   x    n_points    x  n_dims
        # dX_dp:  n_points  x     n_params      x  n_dims
        dW_dp = np.einsum('ild, lpd -> ipd', self.dW_dX, dX_dp)
        # dW_dp:    n_points   x     n_params      x  n_dims

        return dW_dp


# noinspection PyMissingConstructor
class GlobalMDTransform(ModelDrivenTransform):
    r"""
    A transform that couples an alignment transform to a
    statistical model together with a global similarity transform,
    such that the weights of the transform are fully specified by
    both the weights of statistical model and the weights of the
    similarity transform. The model is assumed to
    generate an instance which is then transformed by the similarity
    transform; the result defines the target landmarks of the transform.
    If no source is provided, the mean of the model is defined as the
    source landmarks of the transform.

    Parameters
    ----------
    model : :class:`menpo.model.base.StatisticalModel`
        A linear statistical shape model.
    transform_cls : :class:`menpo.transform.AlignableTransform`
        A class of :class:`menpo.transform.base.AlignableTransform`
        The align constructor will be called on this with the source
        and target landmarks. The target is
        set to the points generated from the model using the
        provide weights - the source is either given or set to the
        model's mean.
    global_transform : :class:`menpo.transform.AlignableTransform`
        A class of :class:`menpo.transform.base.AlignableTransform`
        The global transform that should be applied to the model output.
        Doesn't have to have been constructed from the .align() constructor.
        Note that the GlobalMDTransform isn't guaranteed to hold on to the
        exact object passed in here - so don't expect external changes to
        the global_transform to be reflected in the behavior of this object.
    source : :class:`menpo.shape.base.PointCloud`, optional
        The source landmarks of the transform. If no ``source`` is provided the
        mean of the model is used.
    weights : (P,) ndarray, optional
        The reconstruction weights that will be fed to the model in order to
        generate an instance of the target landmarks.
    composition: 'both', 'warp' or 'model', optional
        The composition approximation employed by this
        ModelDrivenTransform.

        Default: `both`
    """
    def __init__(self, model, transform_cls, global_transform, source=None):
        self.pdm = GlobalPDM(model, global_transform)
        self._cached_points = None
        self.transform = transform_cls(source, self.target)

    def compose_after_from_vector_inplace(self, delta):
        r"""

        Composes two ModelDrivenTransforms together based on the
        first order approximation proposed by Papandreou and Maragos in [1].

        Parameters
        ----------
        delta : (N,) ndarray
            Vectorized :class:`ModelDrivenTransform` to be applied **before**
            self

        Returns
        --------
        transform : self
            self, updated to the result of the composition


        References
        ----------

        .. [1] G. Papandreou and P. Maragos, "Adaptive and Constrained
               Algorithms for Inverse Compositional Active Appearance Model
               Fitting", CVPR08
        """
        model_jacobian = self.pdm.model.jacobian
        points = self.pdm.model.mean.points
        n_points = self.pdm.model.mean.n_points

        # compute:
        # -> dW/dp when p=0
        # -> dW/dp when p!=0
        # -> dW/dx when p!=0 evaluated at the source landmarks

        # dW/dq when p=0 and when p!=0 are the same and given by the
        # Jacobian of the global transform evaluated at the mean of the
        # model
        dW_dq = self._global_transform_jacobian(points)
        # dW_dq:  n_points  x  n_global_params  x  n_dims

        # dW/db when p=0, is the Jacobian of the model
        dW_db_0 = model_jacobian
        # dW_db_0:  n_points  x     n_weights     x  n_dims

        # dW/dp when p=0, is simply the concatenation of the previous
        # two terms
        dW_dp_0 = np.hstack((dW_dq, dW_db_0))
        # dW_dp_0:  n_points  x     n_params      x  n_dims

        # by application of the chain rule dW_db when p!=0,
        # is the Jacobian of the global transform wrt the points times
        # the Jacobian of the model: dX(S)/db = dX/dS *  dS/db
        dW_dS = self.pdm.global_transform.jacobian_points(points)
        dW_db = np.einsum('ilj, idj -> idj', dW_dS, dW_db_0)
        # dW_dS:  n_points  x      n_dims       x  n_dims
        # dW_db:  n_points  x     n_weights     x  n_dims

        # dW/dp is simply the concatenation of dX_dq with dX_db
        dW_dp = np.hstack((dW_dq, dW_db))
        # dW_dp:    n_points  x     n_params     x  n_dims

        dW_dx = self.transform.jacobian_points(points)
        #dW_dx = np.dot(dW_dx, self.global_transform.linear_component.T)
        # dW_dx:  n_points  x  n_dims  x  n_dims

        #TODO: Can we do this without splitting across the two dimensions?
        dW_dx_x = dW_dx[:, 0, :].flatten()[..., None]
        dW_dx_y = dW_dx[:, 1, :].flatten()[..., None]
        dW_dp_0_mat = np.reshape(dW_dp_0, (n_points * self.n_dims,
                                           self.n_parameters))
        dW_dx_dW_dp_0 = dW_dp_0_mat * dW_dx_x + dW_dp_0_mat * dW_dx_y
        dW_dx_dW_dp_0 = np.reshape(dW_dx_dW_dp_0,
                                   (n_points, self.n_parameters, self.n_dims))
        # dW_dx:          n_points  x  n_dims    x  n_dims
        # dW_dp_0:        n_points  x  n_params  x  n_dims
        # dW_dx_dW_dp_0:  n_points  x  n_params  x  n_dims

        J = np.einsum('ijk, ilk -> jl', dW_dp, dW_dx_dW_dp_0)
        H = np.einsum('ijk, ilk -> jl', dW_dp, dW_dp)

        Jp = np.linalg.solve(H, J)
        # Jp:  n_params  x  n_params

        self.from_vector_inplace(self.as_vector() + np.dot(Jp, delta))

    def _global_transform_jacobian(self, points):
        return self.pdm.global_transform.jacobian(points)

    def jacobian(self, points):
        """
        Calculates the Jacobian of the ModelDrivenTransform wrt to
        its weights (the weights). This is done by chaining the relative
        weight of each point wrt the source landmarks, i.e. the Jacobian of
        the warp wrt the source landmarks when the target is assumed to be
        equal to the source (dW/dx), together with the Jacobian of the
        linear model (and of the global transform if present) wrt its
        weights (dX/dp).

        Parameters
        -----------
        points: (N, D) ndarray
            The points at which the Jacobian will be evaluated.

        Returns
        -------
        dW/dp : (N, P, D) ndarray
            The Jacobian of the ModelDrivenTransform evaluated at the
            previous points.
        """
        # check if re-computation of dW/dx can be avoided
        if not np.array_equal(self._cached_points, points):
            # recompute dW/dx, i.e. the relative weight of each point wrt
            # the source landmarks
            self.dW_dX = self.transform.weight_points(points)
            # cache points
            self._cached_points = points

        model_jacobian = self.pdm.model.jacobian
        points = self.pdm.model.mean.points

        # compute dX/dp

        # dX/dq is the Jacobian of the global transform evaluated at the
        # mean of the model.
        dX_dq = self._global_transform_jacobian(points)
        # dX_dq:  n_points  x  n_global_params  x  n_dims

        # by application of the chain rule dX_db is the Jacobian of the
        # model transformed by the linear component of the global transform
        dS_db = model_jacobian
        dX_dS = self.pdm.global_transform.jacobian_points(points)
        dX_db = np.einsum('ilj, idj -> idj', dX_dS, dS_db)
        # dS_db:  n_points  x     n_weights     x  n_dims
        # dX_dS:  n_points  x     n_dims        x  n_dims
        # dX_db:  n_points  x     n_weights     x  n_dims

        # dX/dp is simply the concatenation of the previous two terms
        dX_dp = np.hstack((dX_dq, dX_db))

        # dW_dX:    n_points   x    n_points    x  n_dims
        # dX_dp:  n_points  x     n_params      x  n_dims
        dW_dp = np.einsum('ild, lpd -> ipd', self.dW_dX, dX_dp)
        # dW_dp:    n_points   x     n_params      x  n_dims

        return dW_dp


# noinspection PyMissingConstructor
class OrthoMDTransform(GlobalMDTransform):
    r"""
    A transform that couples an alignment transform to a
    statistical model together with a global similarity transform,
    such that the weights of the transform are fully specified by
    both the weights of statistical model and the weights of the
    similarity transform. The model is assumed to
    generate an instance which is then transformed by the similarity
    transform; the result defines the target landmarks of the transform.
    If no source is provided, the mean of the model is defined as the
    source landmarks of the transform.

    This transform (in contrast to the :class:`GlobalMDTransform`)
    additionally orthonormalizes both the global and the model basis against
    each other, ensuring that orthogonality and normalization is enforced
    across the unified bases.

    Parameters
    ----------
    model : :class:`menpo.model.base.StatisticalModel`
        A linear statistical shape model.
    transform_cls : :class:`menpo.transform.AlignableTransform`
        A class of :class:`menpo.transform.base.AlignableTransform`
        The align constructor will be called on this with the source
        and target landmarks. The target is
        set to the points generated from the model using the
        provide weights - the source is either given or set to the
        model's mean.
    global_transform : :class:`menpo.transform.AlignableTransform`
        A class of :class:`menpo.transform.base.AlignableTransform`
        The global transform that should be applied to the model output.
        Doesn't have to have been constructed from the .align() constructor.
        Note that the GlobalMDTransform isn't guaranteed to hold on to the
        exact object passed in here - so don't expect external changes to
        the global_transform to be reflected in the behavior of this object.
    source : :class:`menpo.shape.base.PointCloud`, optional
        The source landmarks of the transform. If no ``source`` is provided the
        mean of the model is used.
    """
    def __init__(self, model, transform_cls, global_transform, source=None):
        self.pdm = OrthoPDM(model, global_transform)
        self._cached_points = None
        self.transform = transform_cls(source, self.target)

    def _global_transform_jacobian(self, points):
            return self.pdm.similarity_model.jacobian



from __future__ import division
import abc
import numpy as np

from menpo.transform import Scale, Translation
from menpo.groupalign import GeneralizedProcrustesAnalysis
from menpo.model.pca import PCAModel

from .functions import mean_pointcloud


#TODO: Document me
class DeformableModelBuilder(object):
    r"""
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def build(self, images, group=None, label='all'):
        r"""
        """
        pass

    @classmethod
    def _preprocessing(cls, images, group, label, diagonal_range,
                       interpolator, scaled_levels, n_levels, downscale):
        r"""
        """
        print(' - Computing reference shape')
        shapes = [i.landmarks[group][label].lms for i in images]
        reference_shape = mean_pointcloud(shapes)

        print(' - Normalizing object size')
        if diagonal_range:
            x, y = reference_shape.range()
            scale = diagonal_range / np.sqrt(x**2 + y**2)
            Scale(scale, reference_shape.n_dims).apply_inplace(
                reference_shape)
        images = [i.rescale_to_reference_shape(reference_shape,
                                               group=group,
                                               label=label,
                                               interpolator=interpolator)
                  for i in images]

        print(' - Generating multilevel scale space')
        if scaled_levels:
            generator = [i.gaussian_pyramid(n_levels=n_levels,
                                            downscale=downscale)
                         for i in images]
        else:
            generator = [i.smoothing_pyramid(n_levels=n_levels,
                                             downscale=downscale)
                         for i in images]

        return reference_shape, generator

    @classmethod
    def _build_shape_model(cls, shapes, max_components):
        r"""
        """
        # centralize shapes
        centered_shapes = [Translation(-s.centre).apply(s) for s in shapes]
        # align centralized shape using Procrustes Analysis
        gpa = GeneralizedProcrustesAnalysis(centered_shapes)
        aligned_shapes = [s.aligned_source for s in gpa.transforms]

        # build shape model
        shape_model = PCAModel(aligned_shapes)
        if max_components is not None:
            # trim shape model if required
            shape_model.trim_components(max_components)

        return shape_model



from .base import (SupervisedDescentMethodFitter,
                   SupervisedDescentAAMFitter,
                   SupervisedDescentCLMFitter)
from .trainer import (SupervisedDescentMethodTrainer,
                      SupervisedDescentAAMTrainer,
                      SupervisedDescentCLMTrainer)



from menpo.model.pca import PCAModel
from menpo.model.linear import LinearModel, MeanLinearModel
from menpo.model.base import MeanInstanceLinearModel, InstanceLinearModel
from menpo.model.base import Similarity2dInstanceModel



import numpy as np
from numpy.testing import assert_equal
from menpo.transform.piecewiseaffine.base import DiscreteAffinePWA
from menpo.shape import PointCloud, TriMesh

src_points = np.array([[0, 0], [1, 0], [0, 1], [1, 1]])
tgt_points = np.array([[0, 0], [2, 0], [0, 2], [2, 3]])
trilist = np.array([[0, 1, 2], [1, 3, 2]])
src = TriMesh(src_points, trilist)
tgt = PointCloud(tgt_points)

a_affine = np.array(
    [[2.,  0.,  0.],
     [0.,  2.,  0.],
     [0.,  0.,  1.]])

b_affine = np.array(
    [[2.,  0.,  0.],
     [1.,  3., -1.],
     [0.,  0.,  1.]])


def test_pwa_discrete_affine_transforms():
    pwa = DiscreteAffinePWA(src, tgt)
    assert(len(pwa.transforms) == 2)
    assert_equal(pwa.transforms[0].h_matrix, a_affine)
    assert_equal(pwa.transforms[1].h_matrix, b_affine)



# This has to go above the default importers to prevent cyclical importing
import abc

import numpy as np
from scipy.misc import imrotate

from menpo.exception import DimensionalityError
from collections import Iterable


class Renderer(object):
    r"""
    Abstract class for rendering visualizations. Framework specific
    implementations of these classes are made in order to separate
    implementation cleanly from the rest of the code.

    It is assumed that the renderers follow some form of stateful pattern for
    rendering to Figures. Therefore, the major interface for rendering involves
    providing a ``figure_id`` or a boolean about whether a new figure should
    be used. If neither are provided then the default state of the rendering
    engine is assumed to maintained.

    Providing a ``figure_id`` and ``new_figure == True`` is not a valid state.

    Parameters
    ----------
    figure_id : object
        A figure id. Could be any valid object that identifies
        a figure in a given framework (string, int, etc)
    new_figure : bool
        Whether the rendering engine should create a new figure.

    Raises
    ------
    ValueError
        It is not valid to provide a figure id AND request a new figure to
        be rendered on.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, figure_id, new_figure):
        if figure_id is not None and new_figure:
            raise ValueError("Conflicting arguments. figure_id cannot be "
                             "specified if the new_figure flag is True")

        self.figure_id = figure_id
        self.new_figure = new_figure
        self.figure = self.get_figure()

    def render(self, **kwargs):
        r"""
        Render the object on the figure given at instantiation.

        Parameters
        ----------
        kwargs : dict
            Passed through to specific rendering engine.

        Returns
        -------
        viewer : :class:`Renderer`
            Pointer to ``self``.
        """
        return self._render(**kwargs)

    @abc.abstractmethod
    def _render(self, **kwargs):
        r"""
        Abstract method to be overridden the renderer. This will implement the
        actual rendering code for a given object class.

        Parameters
        ----------
        kwargs : dict
            Options to be set when rendering.

        Returns
        -------
        viewer : :class:`Renderer`
            Pointer to ``self``.
        """
        pass

    @abc.abstractmethod
    def get_figure(self):
        r"""
        Abstract method for getting the correct figure to render on. Should
        also set the correct ``figure_id`` for the figure.

        Returns
        -------
        figure : object
            The figure object that the renderer will render on.
        """
        pass


class Viewable(object):
    r"""
    Abstract interface for objects that can visualize themselves.
    """

    __metaclass__ = abc.ABCMeta

    def view_on(self, figure_id, **kwargs):
        r"""
        View the object on a a specific figure specified by the given id.

        Parameters
        ----------
        figure_id : object
            A unique identifier for a figure.
        kwargs : dict
            Passed through to specific rendering engine.

        Returns
        -------
        viewer : :class:`Renderer`
            The renderer instantiated.
        """
        return self._view(figure_id=figure_id, **kwargs)

    def view_new(self, **kwargs):
        r"""
        View the object on a new figure.

        Parameters
        ----------
        kwargs : dict
            Passed through to specific rendering engine.

        Returns
        -------
        viewer : :class:`Renderer`
            The renderer instantiated.
        """
        return self._view(new_figure=True, **kwargs)

    def view(self, **kwargs):
        r"""
        View the object using the default rendering engine figure handling.
        For example, the default behaviour for Matplotlib is that all draw
        commands are applied to the same ``figure`` object.

        Parameters
        ----------
        kwargs : dict
            Passed through to specific rendering engine.

        Returns
        -------
        viewer : :class:`Renderer`
            The renderer instantiated.
        """
        return self._view(**kwargs)

    @abc.abstractmethod
    def _view(self, figure_id=None, new_figure=False, **kwargs):
        r"""
        Abstract method to be overridden by viewable objects. This will
        instantiate a specific visualisation implementation

        Parameters
        ----------
        figure_id : object, optional
            A unique identifier for a figure.

            Default: ``None``
        new_figure : bool, optional
            Whether the rendering engine should create a new figure.

            Default: ``False``
        kwargs : dict
            Passed through to specific rendering engine.

        Returns
        -------
        viewer : :class:`Renderer`
            The renderer instantiated.
        """
        pass


from menpo.visualize.viewmayavi import MayaviPointCloudViewer3d, \
    MayaviTriMeshViewer3d, MayaviTexturedTriMeshViewer3d, \
    MayaviLandmarkViewer3d, MayaviVectorViewer3d, MayaviSurfaceViewer3d, \
    MayaviColouredTriMeshViewer3d
from menpo.visualize.viewmatplotlib import MatplotlibImageViewer2d, \
    MatplotlibImageSubplotsViewer2d, MatplotlibPointCloudViewer2d, \
    MatplotlibLandmarkViewer2d, MatplotlibLandmarkViewer2dImage, \
    MatplotlibTriMeshViewer2d, MatplotlibAlignmentViewer2d, \
    MatplotlibGraphPlotter, MatplotlibMultiImageViewer2d, \
    MatplotlibMultiImageSubplotsViewer2d, MatplotlibFittingViewer2d, \
    MatplotlibFittingSubplotsViewer2d

# Default importer types
PointCloudViewer2d = MatplotlibPointCloudViewer2d
PointCloudViewer3d = MayaviPointCloudViewer3d
TriMeshViewer2d = MatplotlibTriMeshViewer2d
TriMeshViewer3d = MayaviTriMeshViewer3d
TexturedTriMeshViewer3d = MayaviTexturedTriMeshViewer3d
ColouredTriMeshViewer3d = MayaviColouredTriMeshViewer3d
LandmarkViewer3d = MayaviLandmarkViewer3d
LandmarkViewer2d = MatplotlibLandmarkViewer2d
LandmarkViewer2dImage = MatplotlibLandmarkViewer2dImage
ImageViewer2d = MatplotlibImageViewer2d
ImageSubplotsViewer2d = MatplotlibImageSubplotsViewer2d
VectorViewer3d = MayaviVectorViewer3d
DepthImageHeightViewer = MayaviSurfaceViewer3d
AlignmentViewer2d = MatplotlibAlignmentViewer2d
GraphPlotter = MatplotlibGraphPlotter
MultiImageViewer2d = MatplotlibMultiImageViewer2d
MultiImageSubplotsViewer2d = MatplotlibMultiImageSubplotsViewer2d
FittingViewer2d = MatplotlibFittingViewer2d
FittingSubplotsViewer2d = MatplotlibFittingSubplotsViewer2d


class LandmarkViewer(object):
    r"""
    Base Landmark viewer that abstracts away dimensionality

    Parameters
    ----------
    figure_id : object
        A figure id. Could be any valid object that identifies
        a figure in a given framework (string, int, etc)
    new_figure : bool
        Whether the rendering engine should create a new figure.
    group_label : string
        The main label of the landmark set.
    pointcloud : :class:`menpo.shape.pointcloud.PointCloud`
        The pointclouds representing the landmarks.
    labels_to_masks : dict(string, ndarray)
        A dictionary of labels to masks into the pointcloud that represent
        which points belong to the given label.
    target : :class:`menpo.landmarks.base.Landmarkable`
        The parent shape that we are drawing the landmarks for.
    """

    def __init__(self, figure_id, new_figure,
                 group_label, pointcloud, labels_to_masks, target):
        self.pointcloud = pointcloud
        self.group_label = group_label
        self.labels_to_masks = labels_to_masks
        self.target = target
        self.figure_id = figure_id
        self.new_figure = new_figure

    def render(self, **kwargs):
        r"""
        Select the correct type of landmark viewer for the given parent shape.

        Parameters
        ----------
        kwargs : dict
            Passed through to landmark viewer.

        Returns
        -------
        viewer : :class:`Renderer`
            The rendering object.

        Raises
        ------
        DimensionalityError
            Only 2D and 3D viewers are supported.
        """
        if self.pointcloud.n_dims == 2:
            from menpo.image.base import Image

            if isinstance(self.target, Image):
                return LandmarkViewer2dImage(
                    self.figure_id, self.new_figure,
                    self.group_label, self.pointcloud,
                    self.labels_to_masks).render(**kwargs)
            else:
                return LandmarkViewer2d(self.figure_id, self.new_figure,
                                        self.group_label, self.pointcloud,
                                        self.labels_to_masks).render(**kwargs)
        elif self.pointcloud.n_dims == 3:
            return LandmarkViewer3d(self.figure_id, self.new_figure,
                                    self.group_label, self.pointcloud,
                                    self.labels_to_masks).render(**kwargs)
        else:
            raise DimensionalityError("Only 2D and 3D landmarks are "
                                      "currently supported")


class PointCloudViewer(object):
    r"""
    Base PointCloud viewer that abstracts away dimensionality.

    Parameters
    ----------
    figure_id : object
        A figure id. Could be any valid object that identifies
        a figure in a given framework (string, int, etc)
    new_figure : bool
        Whether the rendering engine should create a new figure.
    points : (N, D) ndarray
        The points to render.
    """

    def __init__(self, figure_id, new_figure, points):
        self.figure_id = figure_id
        self.new_figure = new_figure
        self.points = points

    def render(self, **kwargs):
        r"""
        Select the correct type of pointcloud viewer for the given
        pointcloud dimensionality.

        Parameters
        ----------
        kwargs : dict
            Passed through to pointcloud viewer.

        Returns
        -------
        viewer : :class:`Renderer`
            The rendering object.

        Raises
        ------
        DimensionalityError
            Only 2D and 3D viewers are supported.
        """
        if self.points.shape[1] == 2:
            return PointCloudViewer2d(self.figure_id, self.new_figure,
                                      self.points).render(**kwargs)
        elif self.points.shape[1] == 3:
            return PointCloudViewer3d(self.figure_id, self.new_figure,
                                      self.points).render(**kwargs)
        else:
            raise DimensionalityError("Only 2D and 3D pointclouds are "
                                      "currently supported")


class ImageViewer(object):
    r"""
    Base Image viewer that abstracts away dimensionality. It can visualize
    multiple channels of an image in subplots.

    Parameters
    ----------
    figure_id : object
        A figure id. Could be any valid object that identifies
        a figure in a given framework (string, int, etc)
    new_figure : bool
        Whether the rendering engine should create a new figure.
    dimensions : {2, 3} int
        The number of dimensions in the image
    pixels : (N, D) ndarray
        The pixels to render.
    channels: int or list or 'all' or None
        A specific selection of channels to render. The user can choose either
        a single or multiple channels. If all, render all channels in subplot
        mode. If None and channels are less than 36, render them all. If None
        and channels are more than 36, render the first 36.

        Default: None
    mask: (N, D) ndarray
        A boolean mask to be applied to the image. All points outside the
        mask are set to 0.
    """

    def __init__(self, figure_id, new_figure, dimensions, pixels,
                 channels=None, mask=None):
        pixels = pixels.copy()
        self.figure_id = figure_id
        self.new_figure = new_figure
        self.dimensions = dimensions
        pixels, self.use_subplots = \
            self._parse_channels(channels, pixels)
        self.pixels = self._masked_pixels(pixels, mask)

    def _parse_channels(self, channels, pixels):
        r"""
        Parse channels parameter. If channels is int or list, keep it as is. If
        channels is all, return a list of all the image's channels. If channels
        is None, return the minimum between an upper_limit and the image's
        number of channels. If image is grayscale or RGB and channels is None,
        then do not plot channels in different subplots.

        Parameters
        ----------
        channels: int or list or 'all' or None
            A specific selection of channels to render.
        pixels : (N, D) ndarray
            The image's pixels to render.
        upper_limit: int
            The upper limit of subplots for the channels=None case.
        """
        # Flag to trigger ImageSubplotsViewer2d or ImageViewer2d
        use_subplots = True
        n_channels = pixels.shape[2]
        if channels is None:
            if n_channels == 1:
                pixels = pixels[..., 0]
                use_subplots = False
            elif n_channels == 3:
                use_subplots = False
        elif channels != 'all':
            if isinstance(channels, Iterable):
                pixels = pixels[..., channels]
            else:
                pixels = pixels[..., channels]
                use_subplots = False

        return pixels, use_subplots

    def _masked_pixels(self, pixels, mask):
        r"""
        Return the masked pixels using a given boolean mask. In order to make
        sure that the non-masked pixels are visualized in black, their value
        is set to the minimum between min(pixels) and 0.

        Parameters
        ----------
        pixels : (N, D) ndarray
            The image's pixels to render.
        mask: (N, D) ndarray
            A boolean mask to be applied to the image. All points outside the
            mask are set to 0. If mask is None, then the initial pixels are
            returned.
        """
        if mask is not None:
            pixels[~mask] = min(pixels.min(), 0)
        return pixels

    def render(self, **kwargs):
        r"""
        Select the correct type of image viewer for the given
        image dimensionality.

        Parameters
        ----------
        kwargs : dict
            Passed through to image viewer.

        Returns
        -------
        viewer : :class:`Renderer`
            The rendering object.

        Raises
        ------
        DimensionalityError
            Only 2D images are supported.
        """
        if self.dimensions == 2:
            if self.use_subplots:
                return ImageSubplotsViewer2d(self.figure_id, self.new_figure,
                                             self.pixels).render(**kwargs)
            else:
                return ImageViewer2d(self.figure_id, self.new_figure,
                                     self.pixels).render(**kwargs)
        else:
            raise DimensionalityError("Only 2D images are currently supported")


class TriMeshViewer(object):
    r"""
    Base TriMesh viewer that abstracts away dimensionality.

    Parameters
    ----------
    figure_id : object
        A figure id. Could be any valid object that identifies
        a figure in a given framework (string, int, etc)
    new_figure : bool
        Whether the rendering engine should create a new figure.
    points : (N, D) ndarray
        The points to render.
    trilist : (M, 3) ndarray
        The triangulation for the points.
    """

    def __init__(self, figure_id, new_figure, points, trilist):
        self.figure_id = figure_id
        self.new_figure = new_figure
        self.points = points
        self.trilist = trilist

    def render(self, **kwargs):
        r"""
        Select the correct type of trimesh viewer for the given
        trimesh dimensionality.

        Parameters
        ----------
        kwargs : dict
            Passed through to trimesh viewer.

        Returns
        -------
        viewer : :class:`Renderer`
            The rendering object.

        Raises
        ------
        DimensionalityError
            Only 2D and 3D viewers are supported.
        """
        if self.points.shape[1] == 2:
            return TriMeshViewer2d(self.figure_id, self.new_figure,
                                   self.points, self.trilist).render(**kwargs)

        elif self.points.shape[1] == 3:
            return TriMeshViewer3d(self.figure_id, self.new_figure,
                                   self.points, self.trilist).render(**kwargs)
        else:
            raise DimensionalityError("Only 2D and 3D TriMeshes are "
                                      "currently supported")


class MultipleImageViewer(ImageViewer):

    def __init__(self, figure_id, new_figure, dimensions, pixels_list,
                 channels=None, mask=None):
        super(MultipleImageViewer, self).__init__(
            figure_id, new_figure, dimensions, pixels_list[0],
            channels=channels, mask=mask)
        pixels_list = [self._parse_channels(channels, p)[0]
                       for p in pixels_list]
        self.pixels_list = [self._masked_pixels(p, mask)
                            for p in pixels_list]

    def render(self, **kwargs):
        if self.dimensions == 2:
            if self.use_subplots:
                MultiImageSubplotsViewer2d(self.figure_id, self.new_figure,
                                           self.pixels_list).render(**kwargs)
            else:
                return MultiImageViewer2d(self.figure_id, self.new_figure,
                                          self.pixels_list).render(**kwargs)
        else:
            raise DimensionalityError("Only 2D images are currently supported")


class FittingViewer(ImageViewer):

    def __init__(self, figure_id, new_figure, dimensions, pixels,
                 target_list, channels=None, mask=None):
        super(FittingViewer, self).__init__(
            figure_id, new_figure, dimensions, pixels,
            channels=channels, mask=mask)
        self.target_list = target_list

    def render(self, **kwargs):
        if self.dimensions == 2:
            if self.use_subplots:
                FittingSubplotsViewer2d(
                    self.figure_id, self.new_figure, self.pixels,
                    self.target_list).render(**kwargs)
            else:
                return FittingViewer2d(
                    self.figure_id, self.new_figure, self.pixels,
                    self.target_list).render(**kwargs)
        else:
            raise DimensionalityError("Only 2D images are currently supported")



import abc


class Residual(object):
    r"""
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def error(self):
        pass

    @abc.abstractproperty
    def error_derivative(self):
        pass

    @abc.abstractproperty
    def jacobian(self):
        pass

    @abc.abstractproperty
    def hessian(self):
        pass


class SSD(Residual):

    type = 'SSD'

    def error(self):
        raise ValueError("Not implemented")

    def error_derivative(self):
        raise ValueError("Not implemented")

    def jacobian(self):
        raise ValueError("Not implemented")

    def hessian(self):
        raise ValueError("Not implemented")


class Robust(Residual):

    def __init__(self):
        raise ValueError("Not implemented")

    def error(self):
        raise ValueError("Not implemented")

    def error_derivative(self):
        raise ValueError("Not implemented")

    def jacobian(self):
        raise ValueError("Not implemented")

    def hessian(self):
        raise ValueError("Not implemented")

    @abc.abstractmethod
    def _weights(self):
        pass


class Fair(Robust):

    def _weights(self):
        raise ValueError("Not implemented")


class L1L2(Robust):

    def _weights(self):
        raise ValueError("Not implemented")


class GemanMcClure(Robust):

    def _weights(self):
        raise ValueError("Not implemented")


class Cauchy(Robust):

    def _weights(self):
        raise ValueError("Not implemented")


class Welsch(Robust):

    def _weights(self):
        raise ValueError("Not implemented")


class Huber(Robust):

    def _weights(self):
        raise ValueError("Not implemented")


class Turkey(Robust):

    def _weights(self):
        raise ValueError("Not implemented")



from .adaptive import (AdaptiveForwardAdditive,
                       AdaptiveForwardCompositional,
                       AdaptiveInverseCompositional)

from .alternating import (AlternatingForwardAdditive,
                          AlternatingForwardCompositional,
                          AlternatingInverseCompositional)

from .probabilistic import (ProbabilisticForwardAdditive,
                            ProbabilisticForwardCompositional,
                            ProbabilisticInverseCompositional)

from .projectout import (ProjectOutForwardAdditive,
                         ProjectOutForwardCompositional,
                         ProjectOutInverseCompositional)

from .simultaneous import (SimultaneousForwardAdditive,
                           SimultaneousForwardCompositional,
                           SimultaneousInverseCompositional)



import abc

from menpo.fit.base import Fitter
from menpo.fit.fittingresult import (NonParametricFittingResult,
                                     SemiParametricFittingResult,
                                     ParametricFittingResult)


#TODO: document me
class Regressor(Fitter):

    def __init__(self, regressor, features):
        self.regressor = regressor
        self.features = features

    def _set_up(self):
        pass

    def _fit(self, fitting, max_iters=1):
        image = fitting.image
        initial_shape = fitting.initial_shape
        n_iters = 0

        while n_iters < max_iters:
            features = self.features(image, initial_shape)
            delta_p = self.regressor(features)

            fitted_shape, parameters = self.update(delta_p, initial_shape)
            fitting.parameters.append(parameters)
            n_iters += 1

        fitting.fitted = True
        return fitting

    @abc.abstractmethod
    def update(self, delta_p, initial_shape):
        pass


#TODO: document me
class NonParametricRegressor(Regressor):

    def __init__(self, regressor, features):
        super(NonParametricRegressor, self).__init__(
            regressor, features)

    @property
    def algorithm(self):
        return "Non-Parametric"

    def _create_fitting(self, image, shape, gt_shape=None):
        return NonParametricFittingResult(image, self, shape=[shape],
                                          gt_shape=gt_shape)

    def update(self, delta_shape, initial_shape):
        fitted_shape = initial_shape.from_vector(
            initial_shape.as_vector() + delta_shape)
        return fitted_shape, fitted_shape


#TODO: Document me
class SemiParametricRegressor(Regressor):

    def __init__(self, regressor, features, transform, update='composition'):
        super(SemiParametricRegressor, self).__init__(
            regressor, features)
        self.transform = transform
        self._update = self._select_update(update)

    @property
    def algorithm(self):
        return "SemiParametric"

    def _create_fitting(self, image, shape, gt_shape=None):
        self.transform.set_target(shape)
        return SemiParametricFittingResult(
            image, self, parameters=[self.transform.as_vector()],
            gt_shape=gt_shape)

    def _select_update(self, update):
        if update is 'additive':
            return self._additive
        elif update is 'compositional':
            return self._compositional
        else:
            raise ValueError('Unknown update string selected. Valid'
                             'options are: additive, compositional')

    def _additive(self, delta_p):
        parameters = self.transform.as_vector() + delta_p
        self.transform.from_vector_inplace(parameters)

    def _compositional(self, delta_p):
        self.transform.compose_after_from_vector_inplace(delta_p)

    def update(self, delta_p, initial_shape):
        self._update(delta_p)
        return self.transform.target, self.transform.as_vector()


#TODO: Document me
class ParametricRegressor(SemiParametricRegressor):

    def __init__(self, regressor, features, appearance_model, transform,
                 update='composition'):
        super(ParametricRegressor, self).__init__(
            regressor, features, transform, update=update)
        self.appearance_model = appearance_model
        self.template = appearance_model.mean

    @property
    def algorithm(self):
        return "Parametric"

    def _create_fitting(self, image, shape, gt_shape=None):
        self.transform.set_target(shape)
        return ParametricFittingResult(
            image, self, parameters=[self.transform.as_vector()],
            gt_shape=gt_shape)



import numpy as np
import menpo.features as fc


class FeatureExtraction(object):
    r"""
    Class that given an image object, it adds the feature extraction
    functionality to the object and computes the features image. The output
    image's class is either MaskedImage or Image depending on the input image.
    The mask and landmarks are corrected based on the features calculation.

    Parameters
    ----------
    image :  Any image object
        The object created by any image class (Image, MaskedImage,
        SpatialImage, DepthImage).
    """

    def __init__(self, image):
        self._image = image

    def hog(self, mode='dense', algorithm='dalaltriggs', num_bins=9,
            cell_size=8, block_size=2, signed_gradient=True, l2_norm_clip=0.2,
            window_height=1, window_width=1, window_unit='blocks',
            window_step_vertical=1, window_step_horizontal=1,
            window_step_unit='pixels', padding=True, verbose=False,
            constrain_landmarks=True):
        r"""
        Represents a 2-dimensional HOG features image with k number of
        channels. The output object's class is either MaskedImage or Image
        depending on the input image.

        Parameters
        ----------
        mode : 'dense' or 'sparse'
            The 'sparse' case refers to the traditional usage of HOGs, so
            default parameters values are passed to the ImageWindowIterator.
            The sparse case of 'dalaltriggs' algorithm sets the window height
            and width equal to block size and the window step horizontal and
            vertical equal to cell size. Thse sparse case of 'zhuramanan'
            algorithm sets the window height and width equal to 3 times the
            cell size and the window step horizontal and vertical equal to cell
            size. In the 'dense' case, the user can change the
            ImageWindowIterator related parameters (window_height,
            window_width, window_unit, window_step_vertical,
            window_step_horizontal, window_step_unit, padding).

        Default: 'dense'
        window_height : float
            Defines the height of the window for the ImageWindowIterator
            object. The metric unit is defined by window_unit.

            Default: 1
        window_width : float
            Defines the width of the window for the ImageWindowIterator object.
            The metric unit is defined by window_unit.

            Default: 1
        window_unit : 'blocks' or 'pixels'
            Defines the metric unit of the window_height and window_width
            parameters for the ImageWindowIterator object.

            Default: 'blocks'
        window_step_vertical : float
            Defines the vertical step by which the window in the
            ImageWindowIterator is moved, thus it controls the features
            density. The metric unit is defined by window_step_unit.

            Default: 1
        window_step_horizontal : float
            Defines the horizontal step by which the window in the
            ImageWindowIterator is moved, thus it controls the features
            density. The metric unit is defined by window_step_unit.

            Default: 1
        window_step_unit : 'pixels' or 'cells'
            Defines the metric unit of the window_step_vertical and
            window_step_horizontal parameters for the ImageWindowIterator
            object.

            Default: 'pixels'
        padding : bool
            Enables/disables padding for the close-to-boundary windows in the
            ImageWindowIterator object. When padding is enabled, the
            out-of-boundary pixels are set to zero.

            Default: True
        algorithm : 'dalaltriggs' or 'zhuramanan'
            Specifies the algorithm used to compute HOGs.

            Default: 'dalaltriggs'
        cell_size : float
            Defines the cell size in pixels. This value is set to both the
            width and height of the cell. This option is valid for both
            algorithms.

            Default: 8
        block_size : float
            Defines the block size in cells. This value is set to both the
            width and height of the block. This option is valid only for the
            'dalaltriggs' algorithm.

            Default: 2
        num_bins : float
            Defines the number of orientation histogram bins. This option is
            valid only for the 'dalaltriggs' algorithm.

            Default: 9
        signed_gradient : bool
            Flag that defines whether we use signed or unsigned gradient
            angles. This option is valid only for the 'dalaltriggs' algorithm.

            Default: True
        l2_norm_clip : float
            Defines the clipping value of the gradients' L2-norm. This option
            is valid only for the 'dalaltriggs' algorithm.

            Default: 0.2
        constrain_landmarks : bool
            Flag that if enabled, it constrains landmarks that ended up outside
            of the features image bounds.

            Default: True
        verbose : bool
            Flag to print HOG related information.

            Default: False

        Raises
        -------
        ValueError
            HOG features mode must be either dense or sparse
        ValueError
            Algorithm must be either dalaltriggs or zhuramanan
        ValueError
            Number of orientation bins must be > 0
        ValueError
            Cell size (in pixels) must be > 0
        ValueError
            Block size (in cells) must be > 0
        ValueError
            Value for L2-norm clipping must be > 0.0
        ValueError
            Window height must be >= block size and <= image height
        ValueError
            Window width must be >= block size and <= image width
        ValueError
            Window unit must be either pixels or blocks
        ValueError
            Horizontal window step must be > 0
        ValueError
            Vertical window step must be > 0
        ValueError
            Window step unit must be either pixels or cells
        """
        # compute hog features and windows_centres
        hog, window_centres = fc.hog(self._image.pixels, mode=mode,
                                     algorithm=algorithm, cell_size=cell_size,
                                     block_size=block_size, num_bins=num_bins,
                                     signed_gradient=signed_gradient,
                                     l2_norm_clip=l2_norm_clip,
                                     window_height=window_height,
                                     window_width=window_width,
                                     window_unit=window_unit,
                                     window_step_vertical=window_step_vertical,
                                     window_step_horizontal=
                                     window_step_horizontal,
                                     window_step_unit=window_step_unit,
                                     padding=padding, verbose=verbose)
        # create hog image object
        hog_image = self._init_feature_image(hog,
                                             window_centres=window_centres,
                                             constrain_landmarks=
                                             constrain_landmarks)
        # store parameters
        hog_image.hog_parameters = {'mode': mode, 'algorithm': algorithm,
                                    'num_bins': num_bins,
                                    'cell_size': cell_size,
                                    'block_size': block_size,
                                    'signed_gradient': signed_gradient,
                                    'l2_norm_clip': l2_norm_clip,
                                    'window_height': window_height,
                                    'window_width': window_width,
                                    'window_unit': window_unit,
                                    'window_step_vertical':
                                    window_step_vertical,
                                    'window_step_horizontal':
                                    window_step_horizontal,
                                    'window_step_unit': window_step_unit,
                                    'padding': padding,
                                    'original_image_height':
                                    self._image.pixels.shape[0],
                                    'original_image_width':
                                    self._image.pixels.shape[1],
                                    'original_image_channels':
                                    self._image.pixels.shape[2]}
        return hog_image

    def igo(self, double_angles=False, constrain_landmarks=True,
            verbose=False):
        r"""
        Represents a 2-dimensional IGO features image with k=[2,4] number of
        channels. The output object's class is either MaskedImage or Image
        depending on the input image.

        Parameters
        ----------
        image :  class:`Image`
            An image object that contains pixels and mask fields.
        double_angles : bool
            Assume that phi represents the gradient orientations. If this flag
            is disabled, the features image is the concatenation of cos(phi)
            and sin(phi), thus 2 channels. If it is enabled, the features image
            is the concatenation of cos(phi), sin(phi), cos(2*phi), sin(2*phi).

            Default: False
        constrain_landmarks : bool
            Flag that if enabled, it constrains landmarks that ended up outside
            of the features image bounds.

            Default: True
        verbose : bool
            Flag to print IGO related information.

            Default: False

        Raises
        -------
        ValueError
            Image has to be 2D in order to extract IGOs.
        """
        # compute igo features
        igo = fc.igo(self._image.pixels, double_angles=double_angles,
                     verbose=verbose)
        # create igo image object
        igo_image = self._init_feature_image(igo, constrain_landmarks=
                                             constrain_landmarks)
        # store parameters
        igo_image.igo_parameters = {'double_angles': double_angles,
                                    'original_image_height':
                                    self._image.pixels.shape[0],
                                    'original_image_width':
                                    self._image.pixels.shape[1],
                                    'original_image_channels':
                                    self._image.pixels.shape[2]}
        return igo_image

    def es(self, constrain_landmarks=True, verbose=False):
        r"""
        Represents a 2-dimensional Edge Structure (ES) features image with k=2
        number of channels. The output object's class is either MaskedImage or
        Image depending on the input image.

        Parameters
        ----------
        image :  class:`Image`
            An image object that contains pixels and mask fields.
        constrain_landmarks : bool
            Flag that if enabled, it constrains landmarks that ended up outside
            of the features image bounds.

            Default: True
        verbose : bool
            Flag to print IGO related information.

            Default: False

        Raises
        -------
        ValueError
            Image has to be 2D in order to extract ES features.
        """
        # compute es features
        es = fc.es(self._image.pixels, verbose=verbose)
        # create es image object
        es_image = self._init_feature_image(es, constrain_landmarks=
                                            constrain_landmarks)
        # store parameters
        es_image.es_parameters = {'original_image_height':
                                  self._image.pixels.shape[0],
                                  'original_image_width':
                                  self._image.pixels.shape[1],
                                  'original_image_channels':
                                  self._image.pixels.shape[2]}
        return es_image

    def lbp(self, radius=range(1, 5), samples=[8]*4, mapping_type='riu2',
            window_step_vertical=1, window_step_horizontal=1,
            window_step_unit='pixels', padding=True, verbose=False,
            constrain_landmarks=True):
        r"""
        Represents a 2-dimensional LBP features image with k number of
        channels. The output object's class is either MaskedImage or Image
        depending on the input image.

        Parameters
        ----------
        radius : int or list of integers
            It defines the radius of the circle (or circles) at which the
            sampling points will be extracted. The radius (or radii) values
            must be greater than zero. There must be a radius value for each
            samples value, thus they both need to have the same length.

            Default: [1, 2, 3, 4]
        samples : int or list of integers
            It defines the number of sampling points that will be extracted at
            each circle. The samples value (or values) must be greater than
            zero. There must be a samples value for each radius value, thus
            they both need to have the same length.

            Default: [8, 8, 8, 8]
        mapping_type : 'u2' or 'ri' or 'riu2' or 'none'
            It defines the mapping type of the LBP codes. Select 'u2' for
            uniform-2 mapping, 'ri' for rotation-invariant mapping, 'riu2' for
            uniform-2 and rotation-invariant mapping and 'none' to use no
            mapping and only the decimal values instead.

            Default: 'riu2'
        window_step_vertical : float
            Defines the vertical step by which the window in the
            ImageWindowIterator is moved, thus it controls the features
            density. The metric unit is defined by window_step_unit.

            Default: 1
        window_step_horizontal : float
            Defines the horizontal step by which the window in the
            ImageWindowIterator is moved, thus it controls the features
            density. The metric unit is defined by window_step_unit.

            Default: 1
        window_step_unit : 'pixels' or 'window'
            Defines the metric unit of the window_step_vertical and
            window_step_horizontal parameters for the ImageWindowIterator
            object.

            Default: 'pixels'
        padding : bool
            Enables/disables padding for the close-to-boundary windows in the
            ImageWindowIterator object. When padding is enabled, the
            out-of-boundary pixels are set to zero.

            Default: True
        verbose : bool
            Flag to print LBP related information.

            Default: False
        constrain_landmarks : bool
            Flag that if enabled, it constrains landmarks that ended up outside
            of the features image bounds.

            Default: True

        Raises
        -------
        ValueError
            Radius and samples must both be either integers or lists
        ValueError
            Radius and samples must have the same length
        ValueError
            Radius must be > 0
        ValueError
            Radii must be > 0
        ValueError
            Samples must be > 0
        ValueError
            Mapping type must be u2, ri, riu2 or none
        ValueError
            Horizontal window step must be > 0
        ValueError
            Vertical window step must be > 0
        ValueError
            Window step unit must be either pixels or window
        """
        # compute lbp features and windows_centres
        lbp, window_centres = fc.lbp(self._image.pixels, radius=radius,
                                     samples=samples,
                                     mapping_type=mapping_type,
                                     window_step_vertical=window_step_vertical,
                                     window_step_horizontal=
                                     window_step_horizontal,
                                     window_step_unit=window_step_unit,
                                     padding=padding, verbose=verbose)
        # create lbp image object
        lbp_image = self._init_feature_image(lbp,
                                             window_centres=window_centres,
                                             constrain_landmarks=
                                             constrain_landmarks)
        # store parameters
        lbp_image.lbp_parameters = {'radius': radius, 'samples': samples,
                                    'mapping_type': mapping_type,
                                    'window_step_vertical':
                                    window_step_vertical,
                                    'window_step_horizontal':
                                    window_step_horizontal,
                                    'window_step_unit': window_step_unit,
                                    'padding': padding,
                                    'original_image_height':
                                    self._image.pixels.shape[0],
                                    'original_image_width':
                                    self._image.pixels.shape[1],
                                    'original_image_channels':
                                    self._image.pixels.shape[2]}
        return lbp_image

    def _init_feature_image(self, feature_pixels, window_centres=None,
                            constrain_landmarks=True):
        r"""
        Creates a new image object to store the feature_pixels. If the original
        object is of MaskedImage class, then the features object is of
        MaskedImage as well. If the original object is of any other image
        class, the output object is of Image class.

        Parameters
        ----------
        feature_pixels :  ndarray.
            The pixels of the features image.

        window_centres :  ndarray.
            The sampled pixels from where the features were extracted. It has
            size n_rows x n_columns x 2, where window_centres[:, :, 0] are the
            row indices and window_centres[:, :, 1] are the column indices.

        constrain_landmarks : bool
            Flag that if enabled, it constrains landmarks to image bounds.

            Default: True
        """
        from menpo.image import MaskedImage, Image
        if isinstance(self._image, MaskedImage):
            # if we have a MaskedImage object
            feature_image = MaskedImage(feature_pixels)
            # fix mask
            self.transfer_mask(feature_image, window_centres=window_centres)
        else:
            # if we have an Image object
            feature_image = Image(feature_pixels)
        # fix landmarks
        self.transfer_landmarks(feature_image, window_centres=window_centres,
                                constrain_landmarks=constrain_landmarks)
        if window_centres is not None:
            feature_image.window_centres = window_centres
        return feature_image

    def transfer_landmarks(self, target_image, window_centres=None,
                           constrain_landmarks=True):
        r"""
        Transfers its own landmarks to the target_image object after
        appropriately correcting them. The landmarks correction is achieved
        based on the windows_centres of the features object.

        Parameters
        ----------
        target_image :  Either MaskedImage or Image class.
            The target image object that includes the windows_centres.

        window_centres : ndarray, optional
            If set, use these window centres to rescale the landmarks
            appropriately. If None, no scaling is applied.

        constrain_landmarks : bool
            Flag that if enabled, it constrains landmarks to image bounds.

            Default: True
        """
        target_image.landmarks = self._image.landmarks
        if window_centres is not None:
            if target_image.landmarks.has_landmarks:
                for l_group in target_image.landmarks:
                    l = target_image.landmarks[l_group[0]]
                    # find the vertical and horizontal sampling steps
                    step_vertical = window_centres[0, 0, 0]
                    if window_centres.shape[0] > 1:
                        step_vertical = \
                            (window_centres[1, 0, 0] -
                             window_centres[0, 0, 0])
                    step_horizontal = window_centres[0, 0, 1]
                    if window_centres.shape[1] > 1:
                        step_horizontal = \
                            (window_centres[0, 1, 1] -
                             window_centres[0, 0, 1])
                    # convert points by subtracting offset and dividing with
                    # step at each direction
                    l.lms.points[:, 0] = \
                        (l.lms.points[:, 0] -
                         window_centres[:, :, 0].min()) / \
                        step_vertical
                    l.lms.points[:, 1] = \
                        (l.lms.points[:, 1] -
                         window_centres[:, :, 1].min()) / \
                        step_horizontal
        # constrain landmarks to image bounds if asked
        if constrain_landmarks:
            target_image.constrain_landmarks_to_bounds()

    def transfer_mask(self, target_image, window_centres=None):
        r"""
        Transfers its own mask to the target_image object after
        appropriately correcting it. The mask correction is achieved based on
        the windows_centres of the features object.

        Parameters
        ----------
        target_image :  Either MaskedImage or Image class.
            The target image object that includes the windows_centres.

        window_centres : ndarray, optional
            If set, use these window centres to rescale the landmarks
            appropriately. If None, no scaling is applied.
        """
        from menpo.image import BooleanImage
        mask = self._image.mask.mask  # don't want a channel axis!
        if window_centres is not None:
            mask = mask[window_centres[..., 0], window_centres[..., 1]]
        target_image.mask = BooleanImage(mask.copy())



import abc
import copy
import numpy as np

from menpo.exception import DimensionalityError

from .base import Homogeneous, HomogFamilyAlignment


class Affine(Homogeneous):
    r"""
    The base class for all n-dimensional affine transformations. Provides
    methods to break the transform down into it's constituent
    scale/rotation/translation, to view the homogeneous matrix equivalent,
    and to chain this transform with other affine transformations.

    Parameters
    ----------
    h_matrix : (n_dims + 1, n_dims + 1) ndarray
        The homogeneous matrix of the affine transformation.
    """
    def __init__(self, h_matrix):
        Homogeneous.__init__(self, h_matrix)
        # Affine is a little more constrained (only 2D or 3D supported)
        # so run our verification
        Affine.set_h_matrix(self, h_matrix)

    @classmethod
    def identity(cls, n_dims):
        return Affine(np.eye(n_dims + 1))

    @property
    def h_matrix(self):
        return self._h_matrix

    def set_h_matrix(self, value):
        r"""
        Updates the h_matrix, performing sanity checks.

        The Affine h_matrix is limited in what values are allowed. Account
        for them here.
        """
        shape = value.shape
        if len(shape) != 2 and shape[0] != shape[1]:
            raise ValueError("You need to provide a square homogeneous matrix")
        if self.h_matrix is not None:
            # already have a matrix set! The update better be the same size
            if self.n_dims != shape[0] - 1:
                raise DimensionalityError("Trying to update the homogeneous "
                                          "matrix to a different dimension")
        if shape[0] - 1 not in [2, 3]:
            raise DimensionalityError("Affine Transforms can only be 2D or 3D")
        if not (np.allclose(value[-1, :-1], 0) and
                np.allclose(value[-1, -1], 1)):
            raise ValueError("Bottom row must be [0 0 0 1] or [0, 0, 1]")
        self._h_matrix = value.copy()

    @property
    def linear_component(self):
        r"""
        Returns just the linear transform component of this affine
        transform.

        :type: (D, D) ndarray
        """
        return self.h_matrix[:-1, :-1]

    @property
    def translation_component(self):
        r"""
        Returns just the translation component.

        :type: (D,) ndarray
        """
        return self.h_matrix[:-1, -1]

    def decompose(self):
        r"""
        Uses an SVD to decompose this transform into discrete Affine
        Transforms.

        Returns
        -------
        transforms: list of :class`DiscreteAffine` that
            Equivalent to this affine transform, such that:

            ``reduce(lambda x,y: x.chain(y), self.decompose()) == self``
        """
        from .rotation import Rotation
        from .translation import Translation
        from .scale import Scale
        U, S, V = np.linalg.svd(self.linear_component)
        rotation_2 = Rotation(U)
        rotation_1 = Rotation(V)
        scale = Scale(S)
        translation = Translation(self.translation_component)
        return [rotation_1, scale, rotation_2, translation]

    def __eq__(self, other):
        return np.allclose(self.h_matrix, other.h_matrix)

    def __str__(self):
        rep = repr(self) + '\n'
        rep += str(self.h_matrix) + '\n'
        rep += self._transform_str()
        return rep

    def _transform_str(self):
        r"""
        A string representation explaining what this affine transform does.
        Has to be implemented by base classes.

        Returns
        -------
        str : string
            String representation of transform.
        """
        list_str = [t._transform_str() for t in self.decompose()]
        return reduce(lambda x, y: x + '\n' + y, list_str)

    def _apply(self, x, **kwargs):
        r"""
        Applies this transform to a new set of vectors.

        Parameters
        ----------
        x : (N, D) ndarray
            Array to apply this transform to.


        Returns
        -------
        transformed_x : (N, D) ndarray
            The transformed array.
        """
        return np.dot(x, self.linear_component.T) + self.translation_component

    @property
    def n_parameters(self):
        r"""
        ``n_dims * (n_dims + 1)`` parameters - every element of the matrix bar
        the homogeneous part.

        :type: int

        Examples
        --------
        2D Affine: 6 parameters::

            [p1, p3, p5]
            [p2, p4, p6]

        3D Affine: 12 parameters::

            [p1, p4, p7, p10]
            [p2, p5, p8, p11]
            [p3, p6, p9, p12]
        """
        return self.n_dims * (self.n_dims + 1)

    def as_vector(self):
        r"""
        Return the parameters of the transform as a 1D array. These parameters
        are parametrised as deltas from the identity warp. This does not
        include the homogeneous part of the warp. Note that it flattens using
        Fortran ordering, to stay consistent with Matlab.

        **2D**

        ========= ===========================================
        parameter definition
        ========= ===========================================
        p1        Affine parameter
        p2        Affine parameter
        p3        Affine parameter
        p4        Affine parameter
        p5        Translation in ``x``
        p6        Translation in ``y``
        ========= ===========================================

        3D and higher transformations follow a similar format to the 2D case.

        Returns
        -------
        params : (P,) ndarray
            The values that paramaterise the transform.
        """
        params = self.h_matrix - np.eye(self.n_dims + 1)
        return params[:self.n_dims, :].flatten(order='F')

    def from_vector_inplace(self, p):
        r"""
        Updates this Affine in-place from the new parameters. See
        from_vector for details of the parameter format
        """
        h_matrix = None
        if p.shape[0] is 6:  # 2D affine
            h_matrix = np.eye(3)
            h_matrix[:2, :] += p.reshape((2, 3), order='F')
        elif p.shape[0] is 12:  # 3D affine
            h_matrix = np.eye(4)
            h_matrix[:3, :] += p.reshape((3, 4), order='F')
        else:
            ValueError("Only 2D (6 parameters) or 3D (12 parameters) "
                       "homogeneous matrices are supported.")
        self.set_h_matrix(h_matrix)

    @property
    def composes_inplace_with(self):
        return Affine

    def _build_pseudoinverse(self):
        return Affine(np.linalg.inv(self.h_matrix))

    def jacobian(self, points):
        r"""
        Computes the Jacobian of the transform w.r.t the parameters. This is
        constant for affine transforms.

        The Jacobian generated (for 2D) is of the form::

            x 0 y 0 1 0
            0 x 0 y 0 1

        This maintains a parameter order of::

          W(x;p) = [1 + p1  p3      p5] [x]
                   [p2      1 + p4  p6] [y]
                                        [1]

        Parameters
        ----------
        points : (N, D) ndarray
            The set of points to calculate the jacobian for.

        Returns
        -------
        dW_dp : (N, P, D) ndarray
            A (``n_points``, ``n_params``, ``n_dims``) array representing
            the Jacobian of the transform.
        """
        n_points, points_n_dim = points.shape
        if points_n_dim != self.n_dims:
            raise DimensionalityError(
                "Trying to sample jacobian in incorrect dimensions "
                "(transform is {0}D, sampling at {1}D)".format(
                    self.n_dims, points_n_dim))
        # prealloc the jacobian
        jac = np.zeros((n_points, self.n_parameters, self.n_dims))
        # a mask that we can apply at each iteration
        dim_mask = np.eye(self.n_dims, dtype=np.bool)

        for i, s in enumerate(
                range(0, self.n_dims * self.n_dims, self.n_dims)):
            # i is current axis
            # s is slicing offset
            # make a mask for a single points jacobian
            full_mask = np.zeros((self.n_parameters, self.n_dims), dtype=bool)
            # fill the mask in for the ith axis
            full_mask[slice(s, s + self.n_dims)] = dim_mask
            # assign the ith axis points to this mask, broadcasting over all
            # points
            jac[:, full_mask] = points[:, i][..., None]
            # finally, just repeat the same but for the ones at the end
        full_mask = np.zeros((self.n_parameters, self.n_dims), dtype=bool)
        full_mask[slice(s + self.n_dims, s + 2 * self.n_dims)] = dim_mask
        jac[:, full_mask] = 1
        return jac

    def jacobian_points(self, points):
        r"""
        Computes the Jacobian of the transform wrt the points to which
        the transform is applied to. This is constant for affine transforms.

        The Jacobian for a given point (for 2D) is of the form::

            Jx = [(1 + a),     -b  ]
            Jy = [   b,     (1 + a)]
            J =  [Jx, Jy] = [[(1 + a), -b], [b, (1 + a)]]

        where a and b come from:

            W(x;p) = [1 + a   -b      tx] [x]
                     [b       1 + a   ty] [y]
                                          [1]

        Returns
        -------
        dW/dx: dW/dx: (N, D, D) ndarray
            The Jacobian of the transform wrt the points to which the
            transform is applied to.
        """
        return self.linear_component[None, ...]


class AlignmentAffine(Affine, HomogFamilyAlignment):
    r"""
    Constructs an Affine by finding the optimal affine transform to align
    source to target.

    Parameters
    ----------

    source: :class:`menpo.shape.PointCloud`
        The source pointcloud instance used in the alignment

    target: :class:`menpo.shape.PointCloud`
        The target pointcloud instance used in the alignment

    Notes
    -----

    We want to find the optimal transform M which satisfies

        M a = b

    where `a` and `b` are the source and target homogeneous vectors
    respectively.

       (M a)' = b'
       a' M' = b'
       a a' M' = a b'

       `a a'` is of shape `(n_dim + 1, n_dim + 1)` and so can be inverted
       to solve for M.

       This approach is the analytical linear least squares solution to
       the problem at hand. It will have a solution as long as `(a a')`
       is non-singular, which generally means at least 2 corresponding
       points are required.
    """
    def __init__(self, source, target):
        # first, initialize the alignment
        HomogFamilyAlignment.__init__(self, source, target)
        # now, the Affine
        optimal_h = self._build_alignment_h_matrix(source, target)
        Affine.__init__(self, optimal_h)

    @staticmethod
    def _build_alignment_h_matrix(source, target):
        r"""
        Returns the optimal alignment of source to target.
        """
        a = source.h_points
        b = target.h_points
        return np.linalg.solve(np.dot(a, a.T), np.dot(a, b.T)).T

    def set_h_matrix(self, value):
        r"""
        Upon updating the h_matrix we must resync the target.
        """
        Affine.set_h_matrix(self, value)
        # now update the state
        self._sync_target_from_state()

    def from_vector_inplace(self, p):
        r"""
        Updates this Affine in-place from the new parameters. See
        from_vector for details of the parameter format.
        """
        Affine.from_vector_inplace(self, p)
        self._sync_target_from_state()

    def _sync_state_from_target(self):
        optimal_h = self._build_alignment_h_matrix(self.source, self.target)
        # Use the pure Affine setter (so we don't get syncing)
        Affine.set_h_matrix(self, optimal_h)

    def copy_without_alignment(self):
        return Affine(self.h_matrix.copy())


class DiscreteAffine(object):
    r"""
    A discrete Affine transform operation (such as a :meth:`Scale`,
    :class:`Translation` or :meth:`Rotation`). Has to be able to invertable.
    Make sure you inherit from :class:`DiscreteAffine` first,
    for optimal ``decompose()`` behavior.
    """

    __metaclass__ = abc.ABCMeta

    def decompose(self):
        r"""
        A :class:`DiscreteAffine` is already maximally decomposed -
        return a copy of self in a list.

        Returns
        -------
        transform : :class:`DiscreteAffine`
            Deep copy of ``self``.
        """
        return [copy.deepcopy(self)]



import abc
from menpo.base import Vectorizable
from menpo.landmark import Landmarkable
from menpo.transform.base import Transformable
from menpo.visualize.base import Viewable


class Shape(Vectorizable, Landmarkable, Transformable, Viewable):
    """
    Abstract representation of shape. Shapes are vectorizable, viewable,
    landmarkable and transformable. This base class handles transforming
    landmarks when the shape is transformed. Therefore, implementations of
    Shape have to implement the abstract :meth:`_transform_self_inplace` method that
    handles transforming the Shape itself.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(Shape, self).__init__()

    def _transform_inplace(self, transform):
        """
        Transform the landmarks and the shape itself.

        Parameters
        ----------
        transform : func
            A function to transform the spatial data with

        Returns
        -------
        self : ``self``
            A pointer to ``self`` (the result of :meth:`_transform_self_inplace`).
        """
        self.landmarks._transform_inplace(transform)
        return self._transform_self_inplace(transform)

    @abc.abstractmethod
    def _transform_self_inplace(self, transform):
        """
        Implement this method to transform the concrete implementation of a
        shape. This is then called by the Shape's :meth:`_transform_inplace` method,
        which will have updated the landmarks beforehand.

        Parameters
        ----------
        transform : func
            A function to transform the spatial data with

        Returns
        -------
        self : ``self``
            A pointer to ``self``.
        """
        pass



from __future__ import division
import abc
import numpy as np

from menpo.fit.base import Fitter
from menpo.fit.fittingresult import ParametricFittingResult


class LucasKanade(Fitter):
    r"""
    An abstract base class for implementations of Lucas-Kanade [1]_
    type algorithms.

    This is to abstract away optimisation specific functionality such as the
    calculation of the Hessian (which could be derived using a number of
    techniques, including Gauss-Newton and Levenberg-Marquardt).

    Parameters
    ----------
    image : :class:`pybug.image.base.Image`
        The image to perform the alignment upon.

        .. note:: Only the image is expected within the base class because
            different algorithms expect different kinds of template
            (image/model)
    residual : :class:`pybug.lucaskanade.residual.Residual`
        The kind of residual to be calculated. This is used to quantify the
        error between the input image and the reference object.
    transform : :class:`pybug.transform.base.AlignableTransform`
        The transformation type used to warp the image in to the appropriate
        reference frame. This is used by the warping function to calculate
        sub-pixel coordinates of the input image in the reference frame.
    warp : function
        A function that takes 3 arguments,
        ``warp(`` :class:`image <pybug.image.base.Image>`,
        :class:`template <pybug.image.base.Image>`,
        :class:`transform <pybug.transform.base.AlignableTransform>` ``)``
        This function is intended to perform sub-pixel interpolation of the
        pixel locations calculated by transforming the given image into the
        reference frame of the template. Appropriate functions are given in
        :doc:`pybug.interpolation`.
    optimisation : ('GN',) | ('LM', float), optional
        The optimisation technique used to calculate the Hessian approximation.
        Note that for 'LM' the float is used to set the update step.

        Default: 'GN'
    update_step : float, optional
        The update step used when performing a Levenberg-Marquardt
        optimisation.

        Default: 0.001
    eps : float, optional
        The convergence value. When calculating the level of convergence, if
        the norm of the delta parameter updates is less than ``eps``, the
        algorithm is considered to have converged.

        Default: 1**-10

    Notes
    -----
    The type of optimisation technique chosen will determine properties such
    as the convergence rate of the algorithm. The supported optimisation
    techniques are detailed below:

    ===== ==================== ===============================================
    type  full name            hessian approximation
    ===== ==================== ===============================================
    'GN'  Gauss-Newton         :math:`\mathbf{J^T J}`
    'LM'  Levenberg-Marquardt  :math:`\mathbf{J^T J + \lambda\, diag(J^T J)}`
    ===== ==================== ===============================================

    Attributes
    ----------
    transform
    weights
    n_iters

    References
    ----------
    .. [1] Lucas, Bruce D., and Takeo Kanade.
       "An iterative image registration technique with an application to
       stereo vision." IJCAI. Vol. 81. 1981.
    """
    def __init__(self, residual, transform,
                 interpolator='scipy', optimisation=('GN',), eps=10**-10):
        # set basic state for all Lucas Kanade algorithms
        self.transform = transform
        self.residual = residual
        self.eps = eps
        self.interpolator = interpolator
        # select the optimisation approach and warp function
        self._calculate_delta_p = self._select_optimisation(optimisation)

    def _select_optimisation(self, optimisation):
        if optimisation[0] == 'GD':
            self.update_step = optimisation[1]
            self.__e_lm = 0
            return self._gradient_descent
        if optimisation[0] == 'GN':
            return self._gauss_newton_update
        elif optimisation[0] == 'GN_lp':
            self.lp = optimisation[1]
            return self._gauss_newton_lp_update
        elif optimisation[0] == 'LM':
            self.update_step = optimisation[1]
            self.__e_lm = 0
            return self._levenberg_marquardt_update
        else:
            raise ValueError('Unknown optimisation string selected. Valid'
                             'options are: GN, GN_lp, LM')

    def _gradient_descent(self, sd_delta_p):
        raise NotImplementedError("Gradient descent optimization not "
                                  "implemented yet")

    def _gauss_newton_update(self, sd_delta_p):
        return np.linalg.solve(self._H, sd_delta_p)

    def _gauss_newton_lp_update(self, sd_delta_p):
        raise NotImplementedError("Gauss-Newton lp-norm optimization not "
                                  "implemented yet")

    def _levenberg_marquardt_update(self, sd_delta_p):
        LM = np.diagflat(np.diagonal(self._H))
        H_lm = self._H + (self.update_step * LM)

        if self.residual.error < self.__e_lm:
            # Bad step, increase step
            self.update_step *= 10
        else:
            # Good step, decrease step
            self.update_step /= 10
            self.__e_lm = self.residual.error

        return np.linalg.solve(H_lm, sd_delta_p)

    def _set_up(self, **kwargs):
        pass

    def _create_fitting(self, image, parameters, gt_shape=None):
        return ParametricFittingResult(image, self, parameters=[parameters],
                                       gt_shape=gt_shape)

    def get_parameters(self, shape):
        self.transform.set_target(shape)
        return self.transform.as_vector()



from __future__ import division, print_function
import abc
import numpy as np

from menpo.fitmultilevel.functions import (noisy_align, build_sampling_grid,
                                           extract_local_patches)
from menpo.fitmultilevel.featurefunctions import compute_features, sparse_hog
from menpo.fit.fittingresult import (NonParametricFittingResult,
                                     SemiParametricFittingResult,
                                     ParametricFittingResult)

from .base import (NonParametricRegressor, SemiParametricRegressor,
                   ParametricRegressor)
from .parametricfeatures import weights
from .regressionfunctions import regression, mlr


#TODO: document me
class RegressorTrainer(object):
    r"""
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, reference_shape, regression_type=mlr,
                 regression_features=None, noise_std=0.04, rotation=False,
                 n_perturbations=10):
        self.reference_shape = reference_shape
        self.regression_type = regression_type
        self.regression_features = regression_features
        self.rotation = rotation
        self.noise_std = noise_std
        self.n_perturbations = n_perturbations

    def _regression_data(self, images, gt_shapes, perturbed_shapes):

        n_images = len(images)

        features = []
        delta_ps = []
        for j, (i, s, p_shape) in enumerate(zip(images, gt_shapes,
                                                perturbed_shapes)):
            for ps in p_shape:
                features.append(self.features(i, ps))
                delta_ps.append(self.delta_ps(s, ps))
            print(' - {} % '.format(round(100*(j+1)/n_images)), end='\r')

        return np.asarray(features), np.asarray(delta_ps)

    @abc.abstractmethod
    def features(self, image, shape):
        pass

    @abc.abstractmethod
    def delta_ps(self, gt_shape, perturbed_shape):
        pass

    def train(self, images, shapes, perturbed_shapes=None, **kwargs):
        r"""
        """
        n_images = len(images)
        n_shapes = len(shapes)

        print('- generating regression data')
        if n_images != n_shapes:
            raise ValueError("The number of shapes must be equal to "
                             "the number of images.")

        elif not perturbed_shapes:
            perturbed_shapes = self.perturb_shapes(shapes)
            features, delta_ps = self._regression_data(
                images, shapes, perturbed_shapes)

        elif n_images == len(perturbed_shapes):
            features, delta_ps = self._regression_data(
                images, shapes, perturbed_shapes)

        else:
            raise ValueError("The number of perturbed shapes must be "
                             "equal or multiple to the number of images.")

        print('- performing regression')
        regressor = regression(features, delta_ps, self.regression_type,
                               **kwargs)

        print('- computing regression rmse')
        estimated_delta_ps = regressor(features)
        error = np.sqrt(np.mean(np.sum((delta_ps - estimated_delta_ps) ** 2,
                                       axis=1)))
        print(' - error = {}'.format(error))

        return self._build_regressor(regressor, self.features)

    def perturb_shapes(self, gt_shape):
        return [[self._perturb_shape(s) for _ in range(self.n_perturbations)]
                for s in gt_shape]

    def _perturb_shape(self, gt_shape):
        return noisy_align(self.reference_shape, gt_shape,
                           noise_std=self.noise_std
                           ).apply(self.reference_shape)

    @abc.abstractmethod
    def _build_regressor(self, regressor, features):
        pass


#TODO: document me
class NonParametricRegressorTrainer(RegressorTrainer):
    r"""
    """
    def __init__(self, reference_shape, regression_type=mlr,
                 regression_features=sparse_hog, patch_shape=(16, 16),
                 noise_std=0.04, rotation=False, n_perturbations=10):
        super(NonParametricRegressorTrainer, self).__init__(
            reference_shape, regression_type=regression_type,
            regression_features=regression_features, noise_std=noise_std,
            rotation=rotation, n_perturbations=n_perturbations)
        self.patch_shape = patch_shape
        self.sampling_grid = build_sampling_grid(patch_shape)

    def _create_fitting(self, image, shapes, gt_shape=None):
        return NonParametricFittingResult(image, self, shapes=[shapes],
                                          gt_shape=gt_shape)

    def features(self, image, shape):
        patches = extract_local_patches(image, shape, self.sampling_grid)
        features = [compute_features(p, self.regression_features).pixels.ravel()
                    for p in patches]
        return np.hstack((np.asarray(features).ravel(), 1))

    def delta_ps(self, gt_shape, perturbed_shape):
        return (gt_shape.as_vector() -
                perturbed_shape.as_vector())

    def _build_regressor(self, regressor, features):
        return NonParametricRegressor(regressor, features)


#TODO: Document me
class SemiParametricRegressorTrainer(NonParametricRegressorTrainer):
    r"""
    """
    def __init__(self, transform, reference_shape, regression_type=mlr,
                 regression_features=sparse_hog, patch_shape=(16, 16),
                 update='compositional', noise_std=0.04, rotation=False,
                 n_perturbations=10):
        super(SemiParametricRegressorTrainer, self).__init__(
            reference_shape, regression_type=regression_type,
            regression_features=regression_features, patch_shape=patch_shape,
            noise_std=noise_std, rotation=rotation,
            n_perturbations=n_perturbations)
        self.transform = transform
        self.update = update

    @property
    def algorithm(self):
        return "SemiParametric"

    def _create_fitting(self, image, shapes, gt_shape=None):
        return SemiParametricFittingResult(image, self, parameters=[shapes],
                                           gt_shape=gt_shape)

    def delta_ps(self, gt_shape, perturbed_shape):
        self.transform.target = gt_shape
        gt_ps = self.transform.as_vector()
        self.transform.target = perturbed_shape
        perturbed_ps = self.transform.as_vector()
        return gt_ps - perturbed_ps

    def _build_regressor(self, regressor, features, ):
        return SemiParametricRegressor(regressor, features, self.transform,
                                       self.update)


#TODO: Document me
class ParametricRegressorTrainer(RegressorTrainer):
    r"""
    """
    def __init__(self, appearance_model, transform, reference_shape,
                 regression_type=mlr, regression_features=weights,
                 update='compositional', noise_std=0.04, rotation=False,
                 n_perturbations=10, interpolator='scipy'):
        super(ParametricRegressorTrainer, self).__init__(
            reference_shape, regression_type=regression_type,
            regression_features=regression_features, noise_std=noise_std,
            rotation=rotation, n_perturbations=n_perturbations)
        self.appearance_model = appearance_model
        self.template = appearance_model.mean
        self.regression_features = regression_features
        self.transform = transform
        self.update = update
        self.interpolator = interpolator

    def _create_fitting(self, image, shapes, gt_shape=None):
        return ParametricFittingResult(image, self, parameters=[shapes],
                                       gt_shape=gt_shape)

    def features(self, image, shape):
        self.transform.set_target(shape)
        warped_image = image.warp_to(self.template.mask, self.transform,
                                     interpolator=self.interpolator)
        return np.hstack(
            (self.regression_features(self.appearance_model,
                                      warped_image), 1))

    def delta_ps(self, gt_shape, perturbed_shape):
        self.transform.set_target(gt_shape)
        gt_ps = self.transform.as_vector()
        self.transform.set_target(perturbed_shape)
        perturbed_ps = self.transform.as_vector()
        return gt_ps - perturbed_ps

    def _build_regressor(self, regressor, features):
        return ParametricRegressor(
            regressor, features, self.appearance_model, self.transform,
            self.update)


#TODO: Document me
class SemiParametricClassifierBasedRegressorTrainer(
        NonParametricRegressorTrainer):
    r"""
    """
    def __init__(self, classifiers, transform, reference_shape,
                 regression_type=mlr, patch_shape=(16, 16),
                 noise_std=0.04, rotation=False,
                 n_perturbations=10):
        super(SemiParametricClassifierBasedRegressorTrainer, self).__init__(
            reference_shape, regression_type=regression_type,
            patch_shape=patch_shape, noise_std=noise_std, rotation=rotation,
            n_perturbations=n_perturbations)
        self.classifiers = classifiers
        self.transform = transform
        self.update = 'additive'

    def features(self, image, shape):
        patches = extract_local_patches(image, shape, self.sampling_grid)
        features = [clf(np.reshape(p.pixels, (-1, p.n_channels)))
                    for (clf, p) in zip(self.classifiers, patches)]
        return np.hstack((np.asarray(features).ravel(), 1))

    def _create_fitting(self, image, shapes, gt_shape=None):
        return SemiParametricFittingResult(image, self, parameters=[shapes],
                                           gt_shape=gt_shape)

    def delta_ps(self, gt_shape, perturbed_shape):
        self.transform.set_target(gt_shape)
        gt_ps = self.transform.as_vector()
        self.transform.set_target(perturbed_shape)
        perturbed_ps = self.transform.as_vector()
        return gt_ps - perturbed_ps

    def _build_regressor(self, regressor, features, ):
        return SemiParametricRegressor(regressor, features, self.transform,
                                       self.update)



from __future__ import division
import numpy as np

from menpo.transform import Scale
from menpo.visualize.base import GraphPlotter, MultipleImageViewer
from menpo.fit.fittingresult import FittingResult


class MultilevelFittingResult(FittingResult):
    r"""
    Object that holds the state of a MultipleFitter object (to which it is
    linked) after it has fitted a particular image.

    Parameters
    -----------
    image: :class:`menpo.image.masked.MaskedImage`
        The fitted image.

    multiple_fitter: :class:`menpo.fitter.base.Fitter`
        The fitter object used to fitter the image.

    fittings: :class:`menpo.fitter.fittingresult.FittingResult` list
        A list of basic fitting objects.

    affine_correction: :class: `menpo.transforms.affine.Affine`
        An affine transform that maps the result of the top resolution
        fitting level to the space scale of the original image.

    gt_shape: class:`menpo.shape.PointCloud`, optional
        The ground truth shape associated to the image.

        Default: None

    error_type: 'me_norm', 'me' or 'rmse', optional.
        Specifies the way in which the error between the fitted and
        ground truth shapes is to be computed.

        Default: 'me_norm'
    """

    def __init__(self, image, multiple_fitter, fittings, affine_correction,
                 gt_shape=None, error_type='me_norm'):
        self._error_stop = None  # Internal attribute of error_type setter
        self.fittings = fittings
        self._affine_correction = affine_correction
        super(MultilevelFittingResult, self).__init__(
            image, multiple_fitter, gt_shape=gt_shape, error_type=error_type)

    @property
    def n_levels(self):
        r"""
        Returns the number of levels of the fitter object.
        """
        return self.fitter.n_levels

    @property
    def downscale(self):
        r"""
        Returns the downscale factor used by the multiple fitter.
        """
        return self.fitter.downscale

    @property
    def scaled_levels(self):
        r"""
        Returns True if the shape results returned by the basic fittings
        must be scaled.
        """
        return self.fitter.scaled_levels

    @property
    def fitted(self):
        r"""
        Returns the fitted state of each fitting object.
        """
        return [f.fitted for f in self.fittings]

    @FittingResult.error_type.setter
    def error_type(self, error_type):
        r"""
        Sets the error type according to a set of predefined options.
        """
        if error_type is 'me_norm':
            for f in self.fittings:
                f.error_type = error_type
            self._error_stop = 0.1
            self._error_text = 'Point-to-point error normalized by object ' \
                               'size'
        elif error_type is 'me':
            NotImplementedError("erro_type 'me' not implemented yet")
        elif error_type is 'rmse':
            NotImplementedError("error_type 'rmse' not implemented yet")
        else:
            raise ValueError("Unknown error_type string selected. Valid"
                             "options are: 'me_norm', 'me', 'rmse'")
        self._error_type = error_type

    @property
    def n_iters(self):
        r"""
        Returns the total number of iterations used to fitter the image.
        """
        n_iters = 0
        for f in self.fittings:
            n_iters += f.n_iters
        return n_iters

    def shapes(self, as_points=False):
        r"""
        Generates a list containing the shapes obtained at each fitting
        iteration.

        Parameters
        -----------
        as_points: boolean, optional
            Whether the results is returned as a list of PointClouds or
            ndarrays.

            Default: False

        Returns
        -------
        shapes: :class:`menpo.shape.PointCoulds or ndarray list
            A list containing the shapes obtained at each fitting iteration.
        """
        n = self.n_levels - 1

        shapes = []
        for j, f in enumerate(self.fittings):
            if self.scaled_levels:
                transform = Scale(self.downscale**(n-j), 2)
                for t in f.shapes(as_points=as_points):
                    transform.apply_inplace(t)
                    shapes.append(self._affine_correction.apply(t))
            else:
                for t in f.shapes(as_points=as_points):
                    shapes.append(self._affine_correction.apply(t))

        return shapes

    @property
    def final_shape(self):
        r"""
        Returns the final fitted shape.
        """
        return self._affine_correction.apply(
            self.fittings[-1].final_shape)

    @property
    def initial_shape(self):
        r"""
        Returns the initial shape from which the fitting started.
        """
        n = self.n_levels - 1

        initial_shape = self.fittings[0].initial_shape
        if self.scaled_levels:
            Scale(self.downscale ** n,
                  initial_shape.n_dims).apply_inplace(initial_shape)

        return self._affine_correction.apply(initial_shape)

    @FittingResult.gt_shape.setter
    def gt_shape(self, value):
        r"""
        Setter for the ground truth shape associated to the image.
        """
        self._gt_shape = value

    def plot_error(self, figure_id=None, new_figure=False, **kwargs):
        r"""
        Plots the error evolution throughout the fitting.
        """
        if self.gt_shape is not None:
            title = 'Error evolution'
            legend = [self.algorithm]
            x_label = 'Number of iterations'
            y_label = self._error_text
            errors = self.errors
            x_limit = self.n_iters + self.n_levels
            axis_limits = [0, x_limit, 0, np.max(errors)]
            return GraphPlotter(figure_id, new_figure, range(0, x_limit),
                                [errors], title=title, legend=legend,
                                x_label=x_label, y_label=y_label,
                                axis_limits=axis_limits).render(**kwargs)
        else:
            raise ValueError('Ground truth shape has not been set, error '
                             'cannot be plotted')

    # TODO : this should overwrite __str__
    def print_fitting_info(self):
        r"""
        Prints information related to the fitting.
        """
        print "Initial error: {}".format(self.initial_error)
        print "Final error: {}".format(self.final_error)


class AAMMultilevelFittingResult(MultilevelFittingResult):
    r"""
    Object let us recover the state of an AAM Fitter after the latter has
    fitted a particular image.

    Parameters
    -----------
    image: :class:`pybug.image.masked.MaskedImage`
        The fitted image.

    aam_fitter: :class:`pybug.aam.fitter.AAMFitter`
        The aam_fitter object used to fitter the image.

    basic_fittings: :class:`pybug.aam.fitting.BasicFitting` list
        A list of basic fitting objects.

    _affine_correction: :class: `pybug.transforms.affine.Affine`
            An affine transform that maps the result of the top resolution
            fitting level to the space scale of the original image.

    gt_shape: class:`pybug.shape.PointCloud`, optional
        The ground truth shape associated to the image.

        Default: None

    error_type: 'me_norm', 'me' or 'rmse', optional.
        Specifies the way in which the error between the fitted and
        ground truth shapes is to be computed.

        Default: 'me_norm'
    """

    @property
    def residual(self):
        r"""
        Returns the type of residual used by the basic fitter associated to
        each basic fitting.
        """
        # TODO: ensure that all basic_fitting residuals are the same?
        return self.fittings[-1].residual.type

    @property
    def costs(self):
        r"""
        Returns a list containing the cost at each fitting iteration.
        """
        raise ValueError('costs not implemented yet.')
        #return self._flatten_out([f.costs for f in self.basic_fittings])

    @staticmethod
    def _flatten_out(list_of_lists):
        return [i for l in list_of_lists for i in l]

    @property
    def final_cost(self):
        r"""
        Returns the final fitting cost.
        """
        return self.fittings[-1].final_cost

    @property
    def initial_cost(self):
        r"""
        Returns the initial fitting cost.
        """
        return self.fittings[0].initial_cost

    def warped_images(self, from_basic_fittings=False, as_pixels=False):
        r"""
        Generates a list containing the warped images obtained at each fitting
        iteration.

        Parameters
        -----------
        from_basic_fittings: boolean, optional
            If True, the returned transform per iteration is used to warp
            the internal image representation used by each basic fitter.
            If False, the transforms are used to warp original image.

            Default: False

        as_pixels: boolean, optional
            Whether the result is returned as a list of Images or
            ndarrays.

            Default: False

        Returns
        -------
        warped_images: :class:`pybug.image.masked.MaskedImage` or ndarray list
            A list containing the warped images obtained at each fitting
            iteration.
        """
        if from_basic_fittings:
            return self._flatten_out([f.warped_images(as_pixels=as_pixels)
                                      for f in self.fittings])
        else:
            mask = self.fittings[-1].fitter.template.mask
            transform = self.fittings[-1].fitter.transform
            interpolator = self.fittings[-1].fitter.interpolator
            warped_images = []
            for t in self.shapes():
                transform.target = t
                image = self.image.warp_to(mask, transform,
                                           interpolator=interpolator)
                if as_pixels:
                    image = image.pixels

                warped_images.append(image)

        return warped_images

    def appearance_reconstructions(self, as_pixels=False):
        r"""
        Generates a list containing the appearance reconstruction obtained at
        each fitting iteration.

        Parameters
        -----------
        as_pixels: boolean, optional
            Whether the result is returned as a list of Images or
            ndarrays.

            Default: False

        Returns
        -------
        appearance_reconstructions: :class:`pybug.image.masked.MaskedImage`
                                    or ndarray list
            A list containing the appearance reconstructions obtained at each
            fitting iteration.
        """
        return self._flatten_out(
            [f.appearance_reconstructions(as_pixels=as_pixels)
             for f in self.fittings])

    def plot_cost(self, figure_id=None, new_figure=False, **kwargs):
        r"""
        Plots the cost evolution throughout the fitting.
        """
        title = 'Cost evolution'
        legend = self.algorithm
        x_label = 'Number of iterations'
        y_label = 'Normalized cost'
        costs = [c for cost in self.costs for c in cost]
        total_n_iters = self.n_iters + self.n_levels
        axis_limits = [0, total_n_iters, 0, max(costs)]
        return GraphPlotter(figure_id, new_figure,
                            range(0, self.n_iters+self.n_levels), costs,
                            title=title, legend=legend, x_label=x_label,
                            y_label=y_label,
                            axis_limits=axis_limits).render(**kwargs)

    def view_warped_images(self, figure_id=None, new_figure=False,
                           channels=None, from_basic_fittings=False,
                           **kwargs):
        r"""
        Displays the warped images.
        """
        pixels_list = self.warped_images(
            from_basic_fittings=from_basic_fittings, as_pixels=True)
        return self._view_images(pixels_list, figure_id=figure_id,
                                 new_figure=new_figure, channels=channels,
                                 **kwargs)

    def view_appearance_reconstructions(self, figure_id=None,
                                        new_figure=False, channels=None,
                                        **kwargs):
        r"""
        Displays the appearance reconstructions.
        """
        pixels_list = self.appearance_reconstructions(as_pixels=True)
        return self._view_images(pixels_list, figure_id=figure_id,
                                 new_figure=new_figure, channels=channels,
                                 **kwargs)

    def view_error_images(self, figure_id=None, new_figure=False,
                          channels=None, **kwargs):
        r"""
        Displays the error images.
        """
        warped_images = self.warped_images(as_pixels=True)
        appearances = self.appearance_reconstructions(as_pixels=True)
        pixels_list = [a - i for a, i in zip(appearances, warped_images)]
        return self._view_images(pixels_list, figure_id=figure_id,
                                 new_figure=new_figure, channels=channels,
                                 **kwargs)

    def _view_images(self, pixels_list, figure_id=None, new_figure=False,
                     channels=None, **kwargs):
        return MultipleImageViewer(figure_id, new_figure,
                                   self.image.n_dims, pixels_list,
                                   channels=channels).render(**kwargs)

    def print_fitting_info(self):
        super(AAMMultilevelFittingResult, self).print_fitting_info()
        #print "Initial cost: {}".format(self.initial_cost)
        #print "Final cost: {}".format(self.final_cost)



from __future__ import division, print_function
import abc
import numpy as np

from menpo.transform import Scale, AlignmentSimilarity
from menpo.model.pdm import PDM, OrthoPDM
from menpo.transform.modeldriven import ModelDrivenTransform, OrthoMDTransform

from menpo.fit.regression.trainer import (
    NonParametricRegressorTrainer, ParametricRegressorTrainer,
    SemiParametricClassifierBasedRegressorTrainer)
from menpo.fit.regression.regressionfunctions import mlr
from menpo.fit.regression.parametricfeatures import weights
from menpo.fitmultilevel.functions import mean_pointcloud
from menpo.fitmultilevel.featurefunctions import compute_features, sparse_hog

from .base import (SupervisedDescentMethodFitter, SupervisedDescentAAMFitter,
                   SupervisedDescentCLMFitter)


# TODO: document me
class SupervisedDescentTrainer(object):
    r"""
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, regression_type=mlr, regression_features=None,
                 feature_type=None, n_levels=3, downscale=2,
                 scaled_levels=True, noise_std=0.04, rotation=False,
                 n_perturbations=10, interpolator='scipy', **kwargs):
        self.regression_type = regression_type
        self.regression_features = regression_features
        self.feature_type = feature_type
        self.n_levels = n_levels
        self.downscale = downscale
        self.scaled_levels = scaled_levels
        self.noise_std = noise_std
        self.rotation = rotation
        self.n_perturbations = n_perturbations
        self.interpolator = interpolator

    def train(self, images, group=None, label='all', **kwargs):
        r"""
        """
        print('- Computing reference shape')
        self.reference_shape = self._compute_reference_shape(images, group,
                                                             label)

        print('- Normalizing object size')
        self._rescale_reference_shape()
        images = [i.rescale_to_reference_shape(self.reference_shape,
                                               group=group, label=label,
                                               interpolator=self.interpolator)
                  for i in images]

        print('- Generating multilevel scale space')
        if self.scaled_levels:
            # Gaussian pyramid
            generator = [i.gaussian_pyramid(n_levels=self.n_levels,
                                            downscale=self.downscale)
                         for i in images]
        else:
            # Smoothing pyramid
            generator = [i.smoothing_pyramid(n_levels=self.n_levels,
                                             downscale=self.downscale)
                         for i in images]

        print('- Generating multilevel feature space')
        images = []
        for _ in np.arange(self.n_levels):
            images.append([compute_features(g.next(), self.feature_type)
                           for g in generator])
        images.reverse()

        print('- Extracting ground truth shapes')
        gt_shapes = [[i.landmarks[group][label].lms for i in img]
                     for img in images]

        print('- Building regressors')
        regressors = []
        # for each level
        for j, (level_images, level_gt_shapes) in enumerate(zip(images,
                                                                gt_shapes)):
            print(' - Level {}'.format(j))

            trainer = self._set_regressor_trainer(j)

            if j == 0:
                regressor = trainer.train(level_images, level_gt_shapes,
                                          **kwargs)
            else:
                regressor = trainer.train(level_images, level_gt_shapes,
                                          level_shapes, **kwargs)

            print(' - Generating next level data')

            level_shapes = trainer.perturb_shapes(gt_shapes[0])

            regressors.append(regressor)
            count = 0
            total = len(regressors) * len(images[0]) * len(level_shapes[0])
            for k, r in enumerate(regressors):

                test_images = images[k]
                test_gt_shapes = gt_shapes[k]

                fittings = []
                for (i, gt_s, level_s) in zip(test_images, test_gt_shapes,
                                              level_shapes):
                    fitting_sublist = []
                    for ls in level_s:
                        fitting = r.fit(i, ls)
                        fitting.gt_shape = gt_s
                        fitting_sublist.append(fitting)
                        count += 1

                    fittings.append(fitting_sublist)
                    print(' - {} % '.format(round(100*(count+1)/total)),
                          end='\r')

                if self.scaled_levels:
                    level_shapes = [[Scale(self.downscale,
                                           n_dims=self.reference_shape.n_dims
                                           ).apply(f.final_shape)
                                     for f in fitting_sublist]
                                    for fitting_sublist in fittings]
                else:
                    level_shapes = [[f.final_shape for f in fitting_sublist]
                                    for fitting_sublist in fittings]

            mean_error = np.mean(np.array([f.final_error
                                           for fitting_sublist in fittings
                                           for f in fitting_sublist]))
            print(' - Mean error = {}'.format(mean_error))

        return self._build_supervised_descent_fitter(regressors)

    @abc.abstractmethod
    def _compute_reference_shape(self, images, group, label):
        r"""
        """
        pass

    def _rescale_reference_shape(self):
        r"""
        """
        pass

    @abc.abstractmethod
    def _set_regressor_trainer(self, **kwargs):
        r"""
        """
        pass

    @abc.abstractmethod
    def _build_supervised_descent_fitter(self, regressors):
        pass


#TODO: Document me
class SupervisedDescentMethodTrainer(SupervisedDescentTrainer):
    r"""
    """
    def __init__(self, regression_type=mlr, regression_features=sparse_hog,
                 patch_shape=(16, 16), feature_type=None, n_levels=3,
                 downscale=1.5, scaled_levels=True, noise_std=0.04,
                 rotation=False, n_perturbations=10, diagonal_range=None,
                 interpolator='scipy'):
        super(SupervisedDescentMethodTrainer, self).__init__(
            regression_type=regression_type,
            regression_features=regression_features,
            feature_type=feature_type, n_levels=n_levels,
            downscale=downscale, scaled_levels=scaled_levels,
            noise_std=noise_std, rotation=rotation,
            n_perturbations=n_perturbations, interpolator=interpolator)
        self.patch_shape = patch_shape
        self.diagonal_range = diagonal_range

    def _compute_reference_shape(self, images, group, label):
        shapes = [i.landmarks[group][label].lms for i in images]
        return mean_pointcloud(shapes)

    def _rescale_reference_shape(self):
        if self.diagonal_range:
            x, y = self.reference_shape.range()
            scale = self.diagonal_range / np.sqrt(x**2 + y**2)
            Scale(scale, self.reference_shape.n_dims
                  ).apply_inplace(self.reference_shape)

    def _set_regressor_trainer(self, level):
        return NonParametricRegressorTrainer(
            self.reference_shape, regression_type=self.regression_type,
            regression_features=self.regression_features,
            patch_shape=self.patch_shape, noise_std=self.noise_std,
            rotation=self.rotation, n_perturbations=self.n_perturbations)

    def _build_supervised_descent_fitter(self, regressors):
        return SupervisedDescentMethodFitter(
            regressors, self.feature_type, self.reference_shape,
            self.downscale, self.scaled_levels, self.interpolator)


#TODO: Document me
class SupervisedDescentAAMTrainer(SupervisedDescentTrainer):
    r"""
    """
    def __init__(self, aam, regression_type=mlr, regression_features=weights,
                 noise_std=0.04, rotation=False, n_perturbations=10,
                 update='compositional', md_transform=OrthoMDTransform,
                 global_transform=AlignmentSimilarity, n_shape=None,
                 n_appearance=None):
        super(SupervisedDescentAAMTrainer, self).__init__(
            regression_type=regression_type,
            regression_features=regression_features,
            feature_type=aam.feature_type, n_levels=aam.n_levels,
            downscale=aam.downscale, scaled_levels=aam.scaled_levels,
            noise_std=noise_std, rotation=rotation,
            n_perturbations=n_perturbations, interpolator=aam.interpolator)
        self.aam = aam
        self.update = update
        self.md_transform = md_transform
        self.global_transform = global_transform

        if n_shape is not None:
            if type(n_shape) is int:
                for sm in self.aam.shape_models:
                    sm.n_active_components = n_shape
            elif len(n_shape) is 1 and self.aam.n_levels > 1:
                for sm in self.aam.shape_models:
                    sm.n_active_components = n_shape[0]
            elif len(n_shape) is self.aam.n_levels:
                for sm, n in zip(self.aam.shape_models, n_shape):
                    sm.n_active_components = n
            else:
                raise ValueError('n_shape can be integer, integer list '
                                 'containing 1 or {} elements or '
                                 'None'.format(self.aam.n_levels))

        if n_appearance is not None:
            if type(n_appearance) is int:
                for am in self.aam.appearance_models:
                    am.n_active_components = n_appearance
            elif len(n_appearance) is 1 and self.aam.n_levels > 1:
                for am in self.aam.appearance_models:
                    am.n_active_components = n_appearance[0]
            elif len(n_appearance) is self.aam.n_levels:
                for am, n in zip(self.aam.appearance_models, n_shape):
                    am.n_active_components = n
            else:
                raise ValueError('n_appearance can be integer, integer list '
                                 'containing 1 or {} elements or '
                                 'None'.format(self.aam.n_levels))

    def _compute_reference_shape(self, images, group, label):
        return self.aam.reference_shape

    def _normalize_object_size(self, images, group, label):
        return [i.rescale_to_reference_shape(self.reference_shape,
                                             group=group, label=label,
                                             interpolator=self.interpolator)
                for i in images]

    def _set_regressor_trainer(self, level):
        am = self.aam.appearance_models[level]
        sm = self.aam.shape_models[level]

        if self.md_transform is not ModelDrivenTransform:
            md_transform = self.md_transform(
                sm, self.aam.transform, self.global_transform,
                source=am.mean.landmarks['source'].lms)
        else:
            md_transform = self.md_transform(
                sm, self.aam.transform,
                source=am.mean.landmarks['source'].lms)

        return ParametricRegressorTrainer(
            am, md_transform, self.reference_shape,
            regression_type=self.regression_type, regression_features=self
            .regression_features, update=self.update,
            noise_std=self.noise_std, rotation=self.rotation,
            n_perturbations=self.n_perturbations,
            interpolator=self.interpolator)

    def _build_supervised_descent_fitter(self, regressors):
        return SupervisedDescentAAMFitter(self.aam, regressors)


#TODO: Document me
#TODO: Finish me
class SupervisedDescentCLMTrainer(SupervisedDescentTrainer):
    r"""
    """
    def __init__(self, clm, regression_type=mlr, regression_features=weights,
                 noise_std=0.04, rotation=False, n_perturbations=10,
                pdm_transform=OrthoPDM,
                global_transform=AlignmentSimilarity, n_shape=None):
        super(SupervisedDescentCLMTrainer, self).__init__(
            regression_type=regression_type,
            regression_features=regression_features,
            feature_type=clm.feature_type, n_levels=clm.n_levels,
            downscale=clm.downscale, scaled_levels=clm.scaled_levels,
            noise_std=noise_std, rotation=rotation,
            n_perturbations=n_perturbations, interpolator=clm.interpolator)
        self.clm = clm
        self.patch_shape = clm.patch_shape
        self.pdm_transform = pdm_transform
        self.global_transform = global_transform

        if n_shape is not None:
            if type(n_shape) is int:
                for sm in self.clm.shape_models:
                    sm.n_active_components = n_shape
            elif len(n_shape) is 1 and self.clm.n_levels > 1:
                for sm in self.clm.shape_models:
                    sm.n_active_components = n_shape[0]
            elif len(n_shape) is self.clm.n_levels:
                for sm, n in zip(self.clm.shape_models, n_shape):
                    sm.n_active_components = n
            else:
                raise ValueError('n_shape can be integer, integer list '
                                 'containing 1 or {} elements or '
                                 'None'.format(self.clm.n_levels))

    def _compute_reference_shape(self, images, group, label):
        return self.clm.reference_shape

    #TODO: Finish me
    def _set_regressor_trainer(self, level):
        clfs = self.clm.classifiers[level]
        sm = self.clm.shape_models[level]

        if self.pdm_transform is not PDM:
            pdm_transform = self.pdm_transform(sm, self.global_transform)
        else:
            pdm_transform = self.pdm_transform(sm)

        return SemiParametricClassifierBasedRegressorTrainer(
            clfs, pdm_transform, self.reference_shape,
            regression_type=self.regression_type,
            patch_shape=self.patch_shape,
            noise_std=self.noise_std, rotation=self.rotation,
            n_perturbations=self.n_perturbations)

    def _build_supervised_descent_fitter(self, regressors):
        return SupervisedDescentCLMFitter(self.clm, regressors)
